import argparse
import json
import os
import time

import numpy as np
from sklearn.model_selection import train_test_split

from delft.sequenceLabelling import Sequence
from delft.sequenceLabelling.reader import load_data_and_labels_json_offsets
from delft.utilities.Utilities import t_or_f


def configure(
    architecture,
    output_path=None,
    max_sequence_length=-1,
    batch_size=-1,
    embeddings_name=None,
    max_epoch=-1,
    patience=-1,
    early_stop=None,
):
    """
    Set up the default parameters based on the model type.
    """
    model_name = "datasets"

    multiprocessing = True
    o_early_stop = True

    if "BERT" in architecture:
        # architectures with some transformer layer/embeddings inside
        if batch_size == -1:
            # default
            batch_size = 20
        if max_sequence_length == -1:
            # default
            max_sequence_length = 200

        if max_sequence_length > 512:
            # 512 is the largest sequence for BERT input
            max_sequence_length = 512

        embeddings_name = None

    else:
        # RNN-only architectures
        if batch_size == -1:
            batch_size = 20
        if max_sequence_length == -1:
            max_sequence_length = 1500
        multiprocessing = False

    model_name += "-" + architecture

    if batch_size == -1:
        batch_size = 20

    if max_sequence_length == -1:
        max_sequence_length = 3000

    if max_epoch == -1:
        max_epoch = 60

    if early_stop is not None:
        o_early_stop = early_stop

    if patience == -1:
        patience = 5

    return (
        batch_size,
        max_sequence_length,
        model_name,
        embeddings_name,
        max_epoch,
        multiprocessing,
        o_early_stop,
        patience,
    )


# train a model with all available data
def train(
    embeddings_name=None,
    architecture="BidLSTM_CRF",
    transformer=None,
    input_path=None,
    output_path=None,
    fold_count=1,
    features_indices=None,
    max_sequence_length=-1,
    batch_size=-1,
    max_epoch=-1,
    patience=-1,
    learning_rate=None,
    early_stop=None,
    multi_gpu=False,
    num_workers=1,
):

    print("Loading data...")
    if input_path is None:
        x_all1 = y_all1 = x_all2 = y_all2 = x_all3 = y_all3 = []
        dataseer_sentences_path = "data/sequenceLabelling/datasets/dataseer_sentences.json"
        if os.path.exists(dataseer_sentences_path):
            x_all1, y_all1 = load_data_and_labels_json_offsets(dataseer_sentences_path)
        ner_dataset_recognition_sentences_path = (
            "data/sequenceLabelling/datasets/ner_dataset_recognition_sentences.json"
        )
        if os.path.exists(ner_dataset_recognition_sentences_path):
            x_all2, y_all2 = load_data_and_labels_json_offsets(ner_dataset_recognition_sentences_path)
        coleridge_sentences_path = "data/sequenceLabelling/datasets/coleridge_sentences.json.gz"
        if os.path.exists(coleridge_sentences_path):
            x_all3, y_all3 = load_data_and_labels_json_offsets(coleridge_sentences_path)
        x_all = np.concatenate((x_all1, x_all2, x_all3[:1000]), axis=0)
        y_all = np.concatenate((y_all1, y_all2, y_all3[:1000]), axis=0)
    else:
        x_all, y_all = load_data_and_labels_json_offsets(input_path)

    x_train, x_valid, y_train, y_valid = train_test_split(x_all, y_all, test_size=0.1, shuffle=True)

    print(len(x_train), "train sequences")
    print(len(x_valid), "validation sequences")

    batch_size, max_sequence_length, model_name, embeddings_name, max_epoch, multiprocessing, early_stop, patience = (
        configure(
            architecture, output_path, max_sequence_length, batch_size, embeddings_name, max_epoch, patience, early_stop
        )
    )
    model = Sequence(
        model_name,
        recurrent_dropout=0.50,
        embeddings_name=embeddings_name,
        architecture=architecture,
        transformer_name=transformer,
        max_sequence_length=max_sequence_length,
        batch_size=batch_size,
        fold_number=fold_count,
        features_indices=features_indices,
        max_epoch=max_epoch,
        multiprocessing=multiprocessing,
        num_workers=num_workers,
        early_stop=early_stop,
        patience=patience,
        learning_rate=learning_rate,
    )

    start_time = time.time()
    model.train(x_train, y_train, x_valid=x_valid, y_valid=y_valid, multi_gpu=multi_gpu)
    runtime = round(time.time() - start_time, 3)

    print("training runtime: %s seconds " % runtime)

    # saving the model
    if output_path:
        model.save(output_path)
    else:
        model.save()


# split data, train a model and evaluate it
def train_eval(
    embeddings_name=None,
    architecture="BidLSTM_CRF",
    transformer=None,
    input_path=None,
    output_path=None,
    fold_count=1,
    features_indices=None,
    max_sequence_length=-1,
    batch_size=-1,
    max_epoch=-1,
    patience=-1,
    learning_rate=None,
    early_stop=None,
    multi_gpu=False,
    num_workers=1,
):

    print("Loading data...")
    if input_path is None:
        x_all1 = y_all1 = x_all2 = y_all2 = x_all3 = y_all3 = []
        dataseer_sentences_path = "data/sequenceLabelling/datasets/dataseer_sentences.json"
        if os.path.exists(dataseer_sentences_path):
            x_all1, y_all1 = load_data_and_labels_json_offsets(dataseer_sentences_path)
        ner_dataset_recognition_sentences_path = (
            "data/sequenceLabelling/datasets/ner_dataset_recognition_sentences.json"
        )
        if os.path.exists(ner_dataset_recognition_sentences_path):
            x_all2, y_all2 = load_data_and_labels_json_offsets(ner_dataset_recognition_sentences_path)
        coleridge_sentences_path = "data/sequenceLabelling/datasets/coleridge_sentences.json.gz"
        if os.path.exists(coleridge_sentences_path):
            x_all3, y_all3 = load_data_and_labels_json_offsets(coleridge_sentences_path)
        x_all = np.concatenate((x_all1, x_all2, x_all3[:1000]), axis=0)
        y_all = np.concatenate((y_all1, y_all2, y_all3[:1000]), axis=0)
    else:
        x_all, y_all = load_data_and_labels_json_offsets(input_path)

    x_train_all, x_eval, y_train_all, y_eval = train_test_split(x_all, y_all, test_size=0.1, shuffle=True)
    x_train, x_valid, y_train, y_valid = train_test_split(x_train_all, y_train_all, test_size=0.1)

    print(len(x_train), "train sequences")
    print(len(x_valid), "validation sequences")
    print(len(x_eval), "evaluation sequences")

    batch_size, max_sequence_length, model_name, embeddings_name, max_epoch, multiprocessing, early_stop = configure(
        architecture,
        output_path,
        max_sequence_length,
        batch_size,
        embeddings_name,
        max_epoch,
        patience=patience,
        early_stop=early_stop,
    )
    model = Sequence(
        model_name,
        recurrent_dropout=0.50,
        embeddings_name=embeddings_name,
        architecture=architecture,
        transformer_name=transformer,
        max_sequence_length=max_sequence_length,
        batch_size=batch_size,
        fold_number=fold_count,
        features_indices=features_indices,
        max_epoch=max_epoch,
        multiprocessing=multiprocessing,
        num_workers=num_workers,
        early_stop=early_stop,
        patience=patience,
        learning_rate=learning_rate,
    )

    start_time = time.time()

    if fold_count == 1:
        model.train(
            x_train,
            y_train,
            x_valid=x_valid,
            y_valid=y_valid,
        )
    else:
        model.train_nfold(x_train, y_train, x_valid=x_valid, y_valid=y_valid)

    runtime = round(time.time() - start_time, 3)
    print("training runtime: %s seconds " % runtime)

    # evaluation
    print("\nEvaluation:")
    model.eval(x_eval, y_eval)

    # saving the model (must be called after eval for multiple fold training)
    if output_path:
        model.save(output_path)
    else:
        model.save()


def eval_(input_path=None, architecture=None):
    return


# annotate a list of texts
def annotate_text(texts, output_format, architecture="BidLSTM_CRF", features=None, multi_gpu=False):
    annotations = []

    # load model
    model_name = "datasets"
    model_name += "-" + architecture

    model = Sequence(model_name)
    model.load()

    start_time = time.time()

    annotations = model.tag(texts, output_format, features=features, multi_gpu=multi_gpu)
    runtime = round(time.time() - start_time, 3)

    if output_format == "json":
        annotations["runtime"] = runtime
    else:
        print("runtime: %s seconds " % (runtime))

    return annotations


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Trainer for dataset recognition models using the DeLFT library")

    actions = ["train", "train_eval", "eval", "tag"]

    architectures_word_embeddings = [
        "BidLSTM",
        "BidLSTM_CRF",
        "BidLSTM_ChainCRF",
        "BidLSTM_CNN_CRF",
        "BidLSTM_CNN_CRF",
        "BidGRU_CRF",
        "BidLSTM_CNN",
        "BidLSTM_CRF_CASING",
        "BidLSTM_CRF_FEATURES",
        "BidLSTM_ChainCRF_FEATURES",
    ]

    word_embeddings_examples = ["glove-840B", "fasttext-crawl", "word2vec"]

    architectures_transformers_based = [
        "BERT",
        "BERT_CRF",
        "BERT_ChainCRF",
        "BERT_CRF_FEATURES",
        "BERT_CRF_CHAR",
        "BERT_CRF_CHAR_FEATURES",
    ]

    architectures = architectures_word_embeddings + architectures_transformers_based

    pretrained_transformers_examples = ["bert-base-cased", "bert-large-cased", "allenai/scibert_scivocab_cased"]

    parser.add_argument("action", choices=actions)
    parser.add_argument(
        "--fold-count", type=int, default=1, help="Number of fold to use when evaluating with n-fold cross validation."
    )
    parser.add_argument("--architecture", help="Type of model architecture to be used, one of " + str(architectures))
    # group_embeddings = parser.add_mutually_exclusive_group(required=False)
    parser.add_argument(
        "--embedding",
        default=None,
        help="The desired pre-trained word embeddings using their descriptions in the file. "
        + "For local loading, use delft/resources-registry.json. "
        + "Be sure to use here the same name as in the registry, e.g. "
        + str(word_embeddings_examples)
        + " and that the path in the registry to the embedding file is correct on your system.",
    )
    parser.add_argument(
        "--transformer",
        default=None,
        help="The desired pre-trained transformer to be used in the selected architecture. "
        + "For local loading use, delft/resources-registry.json, and be sure to use here the same name as in the registry, e.g. "
        + str(pretrained_transformers_examples)
        + " and that the path in the registry to the model path is correct on your system. "
        + "HuggingFace transformers hub will be used otherwise to fetch the model, see https://huggingface.co/models "
        + "for model names",
    )
    parser.add_argument("--output", help="Directory where to save a trained model.")
    parser.add_argument(
        "--input",
        help="Grobid data file to be used for training (train action), for training and "
        + "evaluation (train_eval action) or just for evaluation (eval action).",
    )
    parser.add_argument("--max-sequence-length", type=int, default=-1, help="max-sequence-length parameter to be used.")
    parser.add_argument("--batch-size", type=int, default=-1, help="batch-size parameter to be used.")
    parser.add_argument(
        "--patience",
        type=int,
        default=-1,
        help="patience, number of extra epochs to perform after the best epoch before stopping a training.",
    )
    parser.add_argument("--learning-rate", type=float, default=None, help="Initial learning rate")

    parser.add_argument("--max-epoch", type=int, default=-1, help="Maximum number of epochs.")
    parser.add_argument(
        "--early-stop",
        type=t_or_f,
        default=None,
        help="Force early training termination when metrics scores are not improving "
        + "after a number of epochs equals to the patience parameter.",
    )
    parser.add_argument(
        "--multi-gpu",
        default=False,
        help="Enable the support for distributed computing (the batch size needs to be set accordingly using --batch-size)",
        action="store_true",
    )

    parser.add_argument(
        "--num-workers",
        type=int,
        default=1,
        help="Number of worker processes for data loading (default: 1, use 0 or 1 for no multiprocessing)",
    )

    args = parser.parse_args()

    action = args.action
    architecture = args.architecture
    output = args.output
    input_path = args.input
    embeddings_name = args.embedding
    max_sequence_length = args.max_sequence_length
    batch_size = args.batch_size
    transformer = args.transformer
    patience = args.patience
    learning_rate = args.learning_rate
    max_epoch = args.max_epoch
    early_stop = args.early_stop
    multi_gpu = args.multi_gpu

    if transformer is None and embeddings_name is None:
        # default word embeddings
        embeddings_name = "glove-840B"

    if action == "train":
        train(
            embeddings_name=embeddings_name,
            architecture=architecture,
            transformer=transformer,
            input_path=input_path,
            output_path=output,
            max_sequence_length=max_sequence_length,
            batch_size=batch_size,
            patience=patience,
            learning_rate=learning_rate,
            max_epoch=max_epoch,
            early_stop=early_stop,
            multi_gpu=multi_gpu,
            num_workers=args.num_workers,
        )

    if action == "eval":
        if args.fold_count is not None and args.fold_count > 1:
            print(
                "The argument fold-count argument will be ignored. For n-fold cross-validation, please use "
                "it in combination with train_eval"
            )
        if input_path is None:
            raise ValueError(
                "A Grobid evaluation data file must be specified to evaluate a grobid model with the parameter --input"
            )
        eval_(input_path=input_path, architecture=architecture)

    if action == "train_eval":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")
        train_eval(
            embeddings_name=embeddings_name,
            architecture=architecture,
            transformer=transformer,
            input_path=input_path,
            output_path=output,
            fold_count=args.fold_count,
            max_sequence_length=max_sequence_length,
            batch_size=batch_size,
            patience=patience,
            learning_rate=learning_rate,
            max_epoch=max_epoch,
            early_stop=early_stop,
            multi_gpu=multi_gpu,
            num_workers=args.num_workers,
        )

    if action == "tag":
        someTexts = []
        someTexts.append(
            "The DEGs were annotated using the following databases: the NR protein database (NCBI), Swiss Prot, Gene Ontology (GO), the Kyoto Encyclopedia of Genes and Genomes (KEGG) database, and the Clusters of Orthologous Groups database (COG) according to the methods of described by Zhou et al"
        )
        someTexts.append(
            "The electrochemiluminescence immunoassay was used to measure serum concentration of 25-hydroxyvitamin D using Roche Modular E170 Analyzer (Roche Diagnostics, Basel, Switzerland)."
        )
        someTexts.append(
            "We found that this technique works very well in practice, for the MNIST and NORB datasets (see below)."
        )
        someTexts.append(
            "We also compare ShanghaiTechRGBD with other RGB-D crowd counting datasets in , and we can see that ShanghaiTechRGBD is the most challenging RGB-D crowd counting dataset in terms of the number of images and heads."
        )
        someTexts.append("Insulin levels of all samples were measured by ELISA kit (Mercodia)")

        result = annotate_text(someTexts, "json", architecture=architecture, multi_gpu=multi_gpu)
        print(json.dumps(result, sort_keys=False, indent=4, ensure_ascii=False))



import argparse
import os
import time

import numpy as np
from sklearn.model_selection import train_test_split

from delft.sequenceLabelling import Sequence
from delft.sequenceLabelling.reader import (
    load_data_and_labels_conll,
    load_data_and_labels_lemonde,
    load_data_and_labels_ontonotes,
)
from delft.utilities.numpy import shuffle_arrays
from delft.utilities.Utilities import stats, t_or_f


def configure(
    architecture,
    dataset_type,
    lang,
    embeddings_name,
    max_sequence_length=-1,
    batch_size=-1,
    patience=-1,
    max_epoch=-1,
    early_stop=None,
):

    o_max_epoch = 60
    o_early_stop = True
    o_multiprocessing = True
    o_max_sequence_length = 300
    o_patience = 5
    o_batch_size = 32

    # general RNN word embeddings input
    if embeddings_name is None:
        o_embeddings_name = "glove-840B"
        if lang == "en":
            if dataset_type == "conll2012":
                o_embeddings_name = "fasttext-crawl"
        elif lang == "fr":
            o_embeddings_name = "wiki.fr"
    else:
        o_embeddings_name = embeddings_name

    if lang == "fr":
        o_multiprocessing = False

    if architecture == "BidLSTM_CNN_CRF":
        o_word_lstm_units = 200
        o_max_epoch = 30
        o_recurrent_dropout = 0.5
    else:
        o_word_lstm_units = 100
        o_max_epoch = 50
        o_recurrent_dropout = 0.5

    # default bert model parameters
    if architecture.find("BERT") != -1:
        o_batch_size = 32
        o_early_stop = True
        o_max_sequence_length = 150
        o_max_epoch = 50
        o_embeddings_name = None

    if dataset_type == "conll2012":
        o_multiprocessing = False

    if patience > 0:
        o_patience = patience

    if batch_size > 0:
        o_batch_size = batch_size

    if max_sequence_length > 0:
        o_max_sequence_length = max_sequence_length

    if max_epoch > 0:
        o_max_epoch = max_epoch

    if early_stop is not None:
        o_early_stop = early_stop

    return (
        o_batch_size,
        o_max_sequence_length,
        o_patience,
        o_recurrent_dropout,
        o_early_stop,
        o_max_epoch,
        o_embeddings_name,
        o_word_lstm_units,
        o_multiprocessing,
    )


# train a model with all available for a given dataset
def train(
    dataset_type="conll2003",
    lang="en",
    embeddings_name=None,
    architecture="BidLSTM_CRF",
    transformer=None,
    data_path=None,
    max_sequence_length=-1,
    batch_size=-1,
    patience=-1,
    learning_rate=None,
    max_epoch=-1,
    early_stop=None,
    multi_gpu=False,
    num_workers=1,
):

    (
        batch_size,
        max_sequence_length,
        patience,
        recurrent_dropout,
        early_stop,
        max_epoch,
        embeddings_name,
        word_lstm_units,
        multiprocessing,
    ) = configure(
        architecture,
        dataset_type,
        lang,
        embeddings_name,
        max_sequence_length,
        batch_size,
        patience,
        max_epoch,
        early_stop,
    )

    if (dataset_type == "conll2003") and (lang == "en"):
        print("Loading data...")
        x_train1, y_train1 = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2003/eng.train")
        x_train2, y_train2 = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2003/eng.testa")
        x_train3, y_train3 = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2003/eng.testb")

        # we concatenate all sets
        x_all = np.concatenate((x_train1, x_train2, x_train3), axis=0)
        y_all = np.concatenate((y_train1, y_train2, y_train3), axis=0)

        # split train and valid sets in a random way
        x_train, x_valid, y_train, y_valid = train_test_split(x_all, y_all, test_size=0.1)
        stats(x_train, y_train, x_valid, y_valid)

        model_name = "ner-en-conll2003-" + architecture

        model = Sequence(
            model_name,
            max_epoch=max_epoch,
            recurrent_dropout=recurrent_dropout,
            embeddings_name=embeddings_name,
            architecture=architecture,
            transformer_name=transformer,
            word_lstm_units=word_lstm_units,
            batch_size=batch_size,
            early_stop=early_stop,
            patience=patience,
            max_sequence_length=max_sequence_length,
            multiprocessing=multiprocessing,
            num_workers=num_workers,
            learning_rate=learning_rate,
        )

    elif (dataset_type == "conll2012") and (lang == "en"):
        print("Loading Ontonotes 5.0 CoNLL-2012 NER data...")

        x_train1, y_train1 = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2012-NER/eng.train")
        x_train2, y_train2 = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2012-NER/eng.dev")
        x_train3, y_train3 = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2012-NER/eng.test")

        # we concatenate train and valid sets
        x_all = np.concatenate((x_train1, x_train2, x_train3), axis=0)
        y_all = np.concatenate((y_train1, y_train2, y_train3), axis=0)

        # split train and valid sets in a random way
        x_train, x_valid, y_train, y_valid = train_test_split(x_all, y_all, test_size=0.1)
        stats(x_train, y_train, x_valid, y_valid)

        model_name = "ner-en-conll2012-" + architecture

        model = Sequence(
            model_name,
            max_epoch=max_epoch,
            recurrent_dropout=recurrent_dropout,
            embeddings_name=embeddings_name,
            architecture=architecture,
            transformer_name=transformer,
            word_lstm_units=word_lstm_units,
            batch_size=batch_size,
            early_stop=early_stop,
            patience=patience,
            max_sequence_length=max_sequence_length,
            multiprocessing=multiprocessing,
            num_workers=num_workers,
            learning_rate=learning_rate,
        )
    elif lang == "fr":
        print("Loading data...")
        dataset_type = "lemonde"
        x_all, y_all = load_data_and_labels_lemonde("data/sequenceLabelling/leMonde/ftb6_ALL.EN.docs.relinked.xml")
        shuffle_arrays([x_all, y_all])
        x_train, x_valid, y_train, y_valid = train_test_split(x_all, y_all, test_size=0.1)
        stats(x_train, y_train, x_valid, y_valid)

        model_name = "ner-fr-lemonde-" + architecture

        model = Sequence(
            model_name,
            max_epoch=max_epoch,
            recurrent_dropout=recurrent_dropout,
            embeddings_name=embeddings_name,
            architecture=architecture,
            transformer_name=transformer,
            word_lstm_units=word_lstm_units,
            batch_size=batch_size,
            early_stop=early_stop,
            patience=patience,
            max_sequence_length=max_sequence_length,
            multiprocessing=multiprocessing,
            num_workers=num_workers,
            learning_rate=learning_rate,
        )
    else:
        print("dataset/language combination is not supported:", dataset_type, lang)
        return

    # elif (dataset_type == 'ontonotes') and (lang == 'en'):
    #    model = sequenceLabelling.Sequence('ner-en-ontonotes', max_epoch=60, embeddings_name=embeddings_name)
    # elif (lang == 'fr'):
    #    model = sequenceLabelling.Sequence('ner-fr-lemonde', max_epoch=60, embeddings_name=embeddings_name)

    start_time = time.time()
    model.train(x_train, y_train, x_valid=x_valid, y_valid=y_valid)
    runtime = round(time.time() - start_time, 3)
    print("training runtime: %s seconds " % (runtime))

    # saving the model
    model.save()


# train and usual eval on dataset, e.g. eval with CoNLL 2003 eng.testb for CoNLL 2003
def train_eval(
    embeddings_name=None,
    dataset_type="conll2003",
    lang="en",
    architecture="BidLSTM_CRF",
    transformer=None,
    fold_count=1,
    train_with_validation_set=False,
    data_path=None,
    patience=-1,
    batch_size=-1,
    max_sequence_length=-1,
    learning_rate=None,
    max_epoch=-1,
    early_stop=None,
    multi_gpu=False,
    num_workers=1,
):

    (
        batch_size,
        max_sequence_length,
        patience,
        recurrent_dropout,
        early_stop,
        max_epoch,
        embeddings_name,
        word_lstm_units,
        multiprocessing,
    ) = configure(
        architecture,
        dataset_type,
        lang,
        embeddings_name,
        max_sequence_length=max_sequence_length,
        batch_size=batch_size,
        patience=patience,
        max_epoch=max_epoch,
        early_stop=early_stop,
    )

    if (dataset_type == "conll2003") and (lang == "en"):
        print("Loading CoNLL 2003 data...")
        x_train, y_train = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2003/eng.train")
        x_valid, y_valid = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2003/eng.testa")
        x_eval, y_eval = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2003/eng.testb")
        stats(x_train, y_train, x_valid, y_valid, x_eval, y_eval)

        model_name = "ner-en-conll2003-" + architecture

        if not train_with_validation_set:
            # restrict training on train set, use validation set for early stop, as in most papers
            model = Sequence(
                model_name,
                max_epoch=max_epoch,
                recurrent_dropout=recurrent_dropout,
                embeddings_name=embeddings_name,
                fold_number=fold_count,
                architecture=architecture,
                transformer_name=transformer,
                word_lstm_units=word_lstm_units,
                batch_size=batch_size,
                early_stop=True,
                patience=patience,
                max_sequence_length=max_sequence_length,
                multiprocessing=multiprocessing,
                learning_rate=learning_rate,
            )
        else:
            # also use validation set to train (no early stop, hyperparmeters must be set preliminarly),
            # as (Chui & Nochols, 2016) and (Peters and al., 2017)
            # this leads obviously to much higher results (~ +0.5 f1 score with CoNLL-2003)
            model = Sequence(
                model_name,
                max_epoch=max_epoch,
                recurrent_dropout=recurrent_dropout,
                embeddings_name=embeddings_name,
                early_stop=False,
                fold_number=fold_count,
                architecture=architecture,
                transformer_name=transformer,
                word_lstm_units=word_lstm_units,
                batch_size=batch_size,
                patience=patience,
                max_sequence_length=max_sequence_length,
                multiprocessing=multiprocessing,
                learning_rate=learning_rate,
            )

    elif (dataset_type == "ontonotes-all") and (lang == "en"):
        print(r"Loading all Ontonotes 5.0 XML data, evaluation will be on 10\% random partition")
        x_all, y_all = load_data_and_labels_ontonotes(data_path)
        x_train_all, x_eval, y_train_all, y_eval = train_test_split(x_all, y_all, test_size=0.1)
        x_train, x_valid, y_train, y_valid = train_test_split(x_train_all, y_train_all, test_size=0.1)
        stats(x_train, y_train, x_valid, y_valid, x_eval, y_eval)

        model_name = "ner-en-ontonotes-" + architecture

        model = Sequence(
            model_name,
            max_epoch=max_epoch,
            recurrent_dropout=recurrent_dropout,
            embeddings_name=embeddings_name,
            fold_number=fold_count,
            architecture=architecture,
            transformer_name=transformer,
            word_lstm_units=word_lstm_units,
            batch_size=batch_size,
            early_stop=early_stop,
            patience=patience,
            max_sequence_length=max_sequence_length,
            multiprocessing=multiprocessing,
            num_workers=num_workers,
            learning_rate=learning_rate,
        )

    elif (dataset_type == "conll2012") and (lang == "en"):
        print("Loading Ontonotes 5.0 CoNLL-2012 NER data...")

        x_train, y_train = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2012-NER/eng.train")
        x_valid, y_valid = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2012-NER/eng.dev")
        x_eval, y_eval = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2012-NER/eng.test")
        stats(x_train, y_train, x_valid, y_valid, x_eval, y_eval)

        model_name = "ner-en-conll2012-" + architecture

        if not train_with_validation_set:
            model = Sequence(
                model_name,
                max_epoch=max_epoch,
                recurrent_dropout=recurrent_dropout,
                embeddings_name=embeddings_name,
                fold_number=fold_count,
                architecture=architecture,
                transformer_name=transformer,
                word_lstm_units=word_lstm_units,
                batch_size=batch_size,
                early_stop=True,
                patience=patience,
                max_sequence_length=max_sequence_length,
                multiprocessing=multiprocessing,
                learning_rate=learning_rate,
            )
        else:
            # also use validation set to train (no early stop, hyperparameters must be set preliminarly),
            # as (Chui & Nochols, 2016) and (Peters and al., 2017)
            # this leads obviously to much higher results
            model = Sequence(
                model_name,
                max_epoch=max_epoch,
                recurrent_dropout=recurrent_dropout,
                embeddings_name=embeddings_name,
                early_stop=False,
                fold_number=fold_count,
                architecture=architecture,
                transformer_name=transformer,
                word_lstm_units=word_lstm_units,
                batch_size=batch_size,
                patience=patience,
                max_sequence_length=max_sequence_length,
                multiprocessing=multiprocessing,
                learning_rate=learning_rate,
            )

    elif (lang == "fr") and (dataset_type == "ftb" or dataset_type is None):
        print("Loading data for ftb...")
        x_all, y_all = load_data_and_labels_lemonde("data/sequenceLabelling/leMonde/ftb6_ALL.EN.docs.relinked.xml")
        shuffle_arrays([x_all, y_all])
        x_train_all, x_eval, y_train_all, y_eval = train_test_split(x_all, y_all, test_size=0.1)
        x_train, x_valid, y_train, y_valid = train_test_split(x_train_all, y_train_all, test_size=0.1)
        stats(x_train, y_train, x_valid, y_valid, x_eval, y_eval)

        model_name = "ner-fr-lemonde-" + architecture

        model = Sequence(
            model_name,
            max_epoch=max_epoch,
            recurrent_dropout=recurrent_dropout,
            embeddings_name=embeddings_name,
            fold_number=fold_count,
            architecture=architecture,
            transformer_name=transformer,
            word_lstm_units=word_lstm_units,
            batch_size=batch_size,
            early_stop=early_stop,
            patience=patience,
            max_sequence_length=max_sequence_length,
            multiprocessing=multiprocessing,
            num_workers=num_workers,
            learning_rate=learning_rate,
        )
    elif (lang == "fr") and (dataset_type == "ftb_force_split"):
        print("Loading data for ftb_force_split...")
        x_train, y_train = load_data_and_labels_conll("data/sequenceLabelling/leMonde/ftb6_train.conll")
        shuffle_arrays([x_train, y_train])
        x_valid, y_valid = load_data_and_labels_conll("data/sequenceLabelling/leMonde/ftb6_dev.conll")
        x_eval, y_eval = load_data_and_labels_conll("data/sequenceLabelling/leMonde/ftb6_test.conll")
        stats(x_train, y_train, x_valid, y_valid, x_eval, y_eval)

        model_name = "ner-fr-lemonde-force-split-" + architecture

        if not train_with_validation_set:
            # restrict training on train set, use validation set for early stop, as in most papers
            model = Sequence(
                model_name,
                max_epoch=max_epoch,
                recurrent_dropout=recurrent_dropout,
                embeddings_name=embeddings_name,
                early_stop=True,
                fold_number=fold_count,
                architecture=architecture,
                transformer_name=transformer,
                word_lstm_units=word_lstm_units,
                batch_size=batch_size,
                patience=patience,
                max_sequence_length=max_sequence_length,
                multiprocessing=multiprocessing,
                learning_rate=learning_rate,
            )
        else:
            # also use validation set to train (no early stop, hyperparmeters must be set preliminarly),
            # as (Chui & Nochols, 2016) and (Peters and al., 2017)
            # this leads obviously to much higher results (~ +0.5 f1 score with CoNLL-2003)
            model = Sequence(
                model_name,
                max_epoch=max_epoch,
                recurrent_dropout=recurrent_dropout,
                embeddings_name=embeddings_name,
                early_stop=False,
                fold_number=fold_count,
                architecture=architecture,
                transformer_name=transformer,
                word_lstm_units=word_lstm_units,
                batch_size=batch_size,
                patience=patience,
                max_sequence_length=max_sequence_length,
                multiprocessing=multiprocessing,
                learning_rate=learning_rate,
            )
    elif (lang == "fr") and (dataset_type == "ftb_force_split_xml"):
        print("Loading data for ftb_force_split_xml...")
        x_train, y_train = load_data_and_labels_lemonde(
            "data/sequenceLabelling/leMonde/ftb6_ALL.EN.docs.relinked.train.xml"
        )
        shuffle_arrays([x_train, y_train])
        x_valid, y_valid = load_data_and_labels_lemonde(
            "data/sequenceLabelling/leMonde/ftb6_ALL.EN.docs.relinked.dev.xml"
        )
        x_eval, y_eval = load_data_and_labels_lemonde(
            "data/sequenceLabelling/leMonde/ftb6_ALL.EN.docs.relinked.test.xml"
        )
        stats(x_train, y_train, x_valid, y_valid, x_eval, y_eval)

        model_name = "ner-fr-lemonde-force-split-xml-" + architecture

        if not train_with_validation_set:
            # restrict training on train set, use validation set for early stop, as in most papers
            model = Sequence(
                model_name,
                max_epoch=max_epoch,
                recurrent_dropout=recurrent_dropout,
                embeddings_name=embeddings_name,
                early_stop=True,
                fold_number=fold_count,
                architecture=architecture,
                transformer_name=transformer,
                word_lstm_units=word_lstm_units,
                batch_size=batch_size,
                patience=patience,
                max_sequence_length=max_sequence_length,
                multiprocessing=multiprocessing,
                learning_rate=learning_rate,
            )
        else:
            # also use validation set to train (no early stop, hyperparmeters must be set preliminarly),
            # as (Chui & Nochols, 2016) and (Peters and al., 2017)
            # this leads obviously to much higher results (~ +0.5 f1 score with CoNLL-2003)
            model = Sequence(
                model_name,
                max_epoch=max_epoch,
                recurrent_dropout=recurrent_dropout,
                embeddings_name=embeddings_name,
                early_stop=False,
                fold_number=fold_count,
                architecture=architecture,
                transformer_name=transformer,
                word_lstm_units=word_lstm_units,
                batch_size=batch_size,
                patience=patience,
                max_sequence_length=max_sequence_length,
                multiprocessing=multiprocessing,
                learning_rate=learning_rate,
            )
    else:
        print("dataset/language combination is not supported:", dataset_type, lang)
        return

    start_time = time.time()
    if fold_count == 1:
        model.train(x_train, y_train, x_valid=x_valid, y_valid=y_valid, multi_gpu=multi_gpu)
    else:
        model.train_nfold(x_train, y_train, x_valid=x_valid, y_valid=y_valid, multi_gpu=multi_gpu)
    runtime = round(time.time() - start_time, 3)
    print("training runtime: %s seconds " % (runtime))

    print("\nEvaluation on test set:")
    model.eval(x_eval, y_eval)

    # # saving the model (must be called after eval for multiple fold training)
    model.save()


# usual eval on CoNLL 2003 eng.testb
def eval(dataset_type="conll2003", lang="en", architecture="BidLSTM_CRF", data_path=None):

    if (dataset_type == "conll2003") and (lang == "en"):
        print("Loading CoNLL-2003 NER data...")
        x_test, y_test = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2003/eng.testb")
        stats(x_eval=x_test, y_eval=y_test)

        # load model
        model_name = "ner-en-conll2003-" + architecture
        model = Sequence(model_name)
        model.load()

    elif (dataset_type == "conll2012") and (lang == "en"):
        print("Loading Ontonotes 5.0 CoNLL-2012 NER data...")

        x_test, y_test = load_data_and_labels_conll("data/sequenceLabelling/CoNLL-2012-NER/eng.test")
        stats(x_eval=x_test, y_eval=y_test)

        # load model
        model_name = "ner-en-conll2012-" + architecture
        model = Sequence(model_name)
        model.load()

    else:
        print("dataset/language combination is not supported for fixed eval:", dataset_type, lang)
        return

    start_time = time.time()

    print("\nEvaluation on test set:")
    model.eval(x_test, y_test)
    runtime = round(time.time() - start_time, 3)

    print("runtime: %s seconds " % (runtime))


# annotate a list of sentences in a file, provides results in a list of offset mentions
def annotate(
    output_format,
    dataset_type="conll2003",
    lang="en",
    architecture="BidLSTM_CRF",
    file_in=None,
    file_out=None,
    multi_gpu=False,
):
    if file_in is None:
        raise ValueError("an input file to be annotated must be provided")

    if not os.path.isfile(file_in):
        raise ValueError("the provided input file is not valid")

    if (dataset_type == "conll2003") and (lang == "en"):
        # load model
        model_name = "ner-en-conll2003-" + architecture
        model = Sequence(model_name)
        model.load()

    elif (dataset_type == "conll2012") and (lang == "en"):
        # load model
        model_name = "ner-en-conll2012-" + architecture
        model = Sequence(model_name)
        model.load()

    elif lang == "fr":
        model_name = "ner-fr-lemonde-" + architecture
        model = Sequence(model_name)
        model.load()
    else:
        print("dataset/language combination is not supported:", dataset_type, lang)
        return

    start_time = time.time()

    model.tag_file(file_in=file_in, output_format=output_format, file_out=file_out, multi_gpu=multi_gpu)
    runtime = round(time.time() - start_time, 3)

    print("runtime: %s seconds " % (runtime))


if __name__ == "__main__":
    architectures_word_embeddings = [
        "BidLSTM",
        "BidLSTM_CRF",
        "BidLSTM_ChainCRF",
        "BidLSTM_CNN_CRF",
        "BidLSTM_CNN_CRF",
        "BidGRU_CRF",
        "BidLSTM_CNN",
        "BidLSTM_CRF_CASING",
    ]

    word_embeddings_examples = ["glove-840B", "fasttext-crawl", "word2vec"]

    architectures_transformers_based = [
        "BERT",
        "BERT_CRF",
        "BERT_ChainCRF",
        "BERT_CRF_FEATURES",
        "BERT_CRF_CHAR",
        "BERT_CRF_CHAR_FEATURES",
    ]

    pretrained_transformers_examples = ["bert-base-cased", "bert-large-cased", "allenai/scibert_scivocab_cased"]

    architectures = architectures_word_embeddings + architectures_transformers_based

    parser = argparse.ArgumentParser(description="Neural Named Entity Recognizers based on DeLFT")

    parser.add_argument("action", help="one of [train, train_eval, eval, tag]")
    parser.add_argument("--fold-count", type=int, default=1, help="number of folds or re-runs to be used when training")
    parser.add_argument("--lang", default="en", help="language of the model as ISO 639-1 code (en, fr, de, etc.)")
    parser.add_argument("--dataset-type", default="conll2003", help="dataset to be used for training the model")
    parser.add_argument(
        "--train-with-validation-set",
        action="store_true",
        help="Use the validation set for training together with the training set",
    )
    parser.add_argument(
        "--architecture",
        default="BidLSTM_CRF",
        help="type of model architecture to be used, one of " + str(architectures),
    )
    parser.add_argument(
        "--data-path",
        default=None,
        help="path to the corpus of documents for training (only use currently with Ontonotes corpus in orginal XML format)",
    )
    parser.add_argument("--file-in", default=None, help="path to a text file to annotate")
    parser.add_argument("--file-out", default=None, help="path for outputting the resulting JSON NER annotations")

    parser.add_argument(
        "--embedding",
        default=None,
        help="The desired pre-trained word embeddings using their descriptions in the file. "
        + "For local loading, use delft/resources-registry.json. "
        + "Be sure to use here the same name as in the registry, e.g. "
        + str(word_embeddings_examples)
        + " and that the path in the registry to the embedding file is correct on your system.",
    )
    parser.add_argument(
        "--transformer",
        default=None,
        help="The desired pre-trained transformer to be used in the selected architecture. "
        + "For local loading use, delft/resources-registry.json, and be sure to use here the same name as in the registry, e.g. "
        + str(pretrained_transformers_examples)
        + " and that the path in the registry to the model path is correct on your system. "
        + "HuggingFace transformers hub will be used otherwise to fetch the model, see https://huggingface.co/models "
        + "for model names",
    )

    parser.add_argument("--max-sequence-length", type=int, default=-1, help="max-sequence-length parameter to be used.")
    parser.add_argument("--batch-size", type=int, default=-1, help="batch-size parameter to be used.")
    parser.add_argument(
        "--patience",
        type=int,
        default=-1,
        help="patience, number of extra epochs to perform after the best epoch before stopping a training.",
    )
    parser.add_argument("--learning-rate", type=float, default=None, help="Initial learning rate")

    parser.add_argument("--max-epoch", type=int, default=-1, help="Maximum number of epochs.")
    parser.add_argument(
        "--early-stop",
        type=t_or_f,
        default=None,
        help="Force early training termination when metrics scores are not improving "
        + "after a number of epochs equals to the patience parameter.",
    )

    parser.add_argument(
        "--multi-gpu",
        default=False,
        help="Enable the support for distributed computing (the batch size needs to be set accordingly using --batch-size)",
        action="store_true",
    )

    parser.add_argument(
        "--num-workers", type=int, default=1, help="Number of worker processes for data loading (default: 1)"
    )

    args = parser.parse_args()

    action = args.action
    if action not in ("train", "tag", "eval", "train_eval"):
        print("action not specified, must be one of [train, train_eval, eval, tag]")
    lang = args.lang
    dataset_type = args.dataset_type
    train_with_validation_set = args.train_with_validation_set
    architecture = args.architecture
    if architecture not in architectures:
        print("unknown model architecture, must be one of", architectures)
    transformer = args.transformer
    data_path = args.data_path
    file_in = args.file_in
    file_out = args.file_out
    patience = args.patience
    max_sequence_length = args.max_sequence_length
    batch_size = args.batch_size
    learning_rate = args.learning_rate
    max_epoch = args.max_epoch
    early_stop = args.early_stop
    multi_gpu = args.multi_gpu

    # name of embeddings refers to the file delft/resources-registry.json
    # be sure to use here the same name as in the registry ('glove-840B', 'fasttext-crawl', 'word2vec'),
    # and that the path in the registry to the embedding file is correct on your system
    # below we set the default embeddings value
    embeddings_name = args.embedding

    if action == "train":
        train(
            dataset_type=dataset_type,
            lang=lang,
            embeddings_name=embeddings_name,
            architecture=architecture,
            transformer=transformer,
            data_path=data_path,
            max_sequence_length=max_sequence_length,
            batch_size=batch_size,
            patience=patience,
            learning_rate=learning_rate,
            max_epoch=max_epoch,
            early_stop=early_stop,
            multi_gpu=multi_gpu,
            num_workers=args.num_workers,
        )

    if action == "train_eval":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")
        train_eval(
            dataset_type=dataset_type,
            lang=lang,
            embeddings_name=embeddings_name,
            architecture=architecture,
            transformer=transformer,
            fold_count=args.fold_count,
            train_with_validation_set=train_with_validation_set,
            data_path=data_path,
            max_sequence_length=max_sequence_length,
            batch_size=batch_size,
            patience=patience,
            learning_rate=learning_rate,
            max_epoch=max_epoch,
            early_stop=early_stop,
            multi_gpu=multi_gpu,
            num_workers=args.num_workers,
        )

    if action == "eval":
        eval(dataset_type=dataset_type, lang=lang, architecture=architecture)

    if action == "tag":
        if lang != "en" and lang != "fr":
            print("Language not supported:", lang)
        else:
            print(file_in)
            annotate(
                "json",
                dataset_type,
                lang,
                architecture=architecture,
                file_in=file_in,
                file_out=file_out,
                multi_gpu=multi_gpu,
            )
            """
            if result is not None:
                if file_out is None:
                    print(json.dumps(result, sort_keys=False, indent=4, ensure_ascii=False))
            """



import argparse
import json
import sys
import time

import pandas as pd

from delft.textClassification import Classifier
from delft.textClassification.models import architectures
from delft.textClassification.reader import load_texts_and_classes_pandas, load_texts_pandas

list_classes = ["toxic", "severe_toxic", "obscene", "threat", "insult", "identity_hate"]
class_weights = {0: 1.0, 1: 1.0, 2: 1.0, 3: 1.0, 4: 1.0, 5: 1.0}


def configure(architecture):
    batch_size = 256
    maxlen = 300
    patience = 5
    early_stop = True
    max_epoch = 30

    # default bert model parameters
    if architecture == "bert":
        batch_size = 32
        early_stop = False
        max_epoch = 3
        maxlen = 200

    return batch_size, maxlen, patience, early_stop, max_epoch


def train(embeddings_name=None, fold_count=1, architecture="gru", transformer=None):
    batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

    model = Classifier(
        "toxic_" + architecture,
        architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        class_weights=class_weights,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        patience=patience,
        early_stop=early_stop,
        transformer_name=transformer,
    )

    print("loading train dataset...")
    xtr, y = load_texts_and_classes_pandas("data/textClassification/toxic/train.csv")
    if fold_count == 1:
        model.train(xtr, y)
    else:
        model.train_nfold(xtr, y)
    # saving the model
    model.save()


def test(architecture="gru"):
    # load model
    model = Classifier("toxic_" + architecture)
    model.load()

    print("loading test dataset...")
    xte = load_texts_pandas("data/textClassification/toxic/test.csv")
    print("number of texts to classify:", len(xte))
    start_time = time.time()
    result = model.predict(xte, output_format="csv")
    print("runtime: %s seconds " % (round(time.time() - start_time, 3)))
    return result


# classify a list of texts
def classify(texts, output_format, architecture="gru", transformer=None):
    # load model
    model = Classifier("toxic_" + architecture)
    model.load()
    start_time = time.time()
    result = model.predict(texts, output_format)
    print("runtime: %s seconds " % (round(time.time() - start_time, 3)))
    return result


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Classification of comments/short texts in toxicity types (toxic, severe_toxic, obscene, threat, insult, identity_hate) based on DeLFT"
    )

    word_embeddings_examples = ["glove-840B", "fasttext-crawl", "word2vec"]
    pretrained_transformers_examples = ["bert-base-cased", "bert-large-cased", "allenai/scibert_scivocab_cased"]

    parser.add_argument("action", help="one of [train, test, classify]")
    parser.add_argument("--fold-count", type=int, default=1)
    parser.add_argument(
        "--architecture", default="gru", help="type of model architecture to be used, one of " + str(architectures)
    )
    parser.add_argument(
        "--embedding",
        default=None,
        help="The desired pre-trained word embeddings using their descriptions in the file. "
        + "For local loading, use delft/resources-registry.json. "
        + "Be sure to use here the same name as in the registry, e.g. "
        + str(word_embeddings_examples)
        + " and that the path in the registry to the embedding file is correct on your system.",
    )
    parser.add_argument(
        "--transformer",
        default=None,
        help="The desired pre-trained transformer to be used in the selected architecture. "
        + "For local loading use, delft/resources-registry.json, and be sure to use here the same name as in the registry, e.g. "
        + str(pretrained_transformers_examples)
        + " and that the path in the registry to the model path is correct on your system. "
        + "HuggingFace transformers hub will be used otherwise to fetch the model, see https://huggingface.co/models "
        + "for model names",
    )

    args = parser.parse_args()

    action = args.action
    if action not in ("train", "classify", "test"):
        print("action not specifed, must be one of [train,test,classify]")

    embeddings_name = args.embedding
    transformer = args.transformer

    architecture = args.architecture
    if architecture not in architectures:
        print("unknown model architecture, must be one of " + str(architectures))

    if transformer is None and embeddings_name is None:
        # default word embeddings
        embeddings_name = "glove-840B"

    if architecture.find("bert") != -1:
        print(
            "BERT models are not supported for multi-label labelling, at least for the moment. Please choose a RNN architecture."
        )
        sys.exit(0)

    if action == "train":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")
        train(
            embeddings_name=embeddings_name,
            fold_count=args.fold_count,
            architecture=architecture,
            transformer=transformer,
        )

    if action == "test":
        y_test = test()

        # write test predictions as a submission file
        sample_submission = pd.read_csv("data/textClassification/toxic/sample_submission.csv")
        sample_submission[list_classes] = y_test
        sample_submission.to_csv("data/textClassification/toxic/result.csv", index=False)

    if action == "classify":
        someTexts = [
            "This is a gentle test.",
            "This is a fucking test!",
            "With all due respects, I think you're a moron.",
        ]
        result = classify(someTexts, "json", architecture=architecture)
        print(json.dumps(result, sort_keys=False, indent=4, ensure_ascii=False))



import os

# for using legacy Keras 2, and not Keras 3 installed by default from TensorFlow 2.16
os.environ["TF_USE_LEGACY_KERAS"] = "1"
os.environ["KERAS_BACKEND"] = "tensorflow"

# LF: do not remove this import. This ensures that CUDA 12 from tensorflow
#    get initialised before CUDA 11 from pytorch.
import tf_keras as keras  # noqa: F401

from delft.sequenceLabelling.trainer import LogLearningRateCallback

# ask tensorflow to be quiet and not print hundred lines of logs
from delft.utilities.misc import print_parameters

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
os.environ["TOKENIZERS_PARALLELISM"] = "false"

import warnings

import numpy as np

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)

import tensorflow as tf

tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

from tensorflow.python.util import deprecation

deprecation._PRINT_DEPRECATION_WARNINGS = False

# unfortunately when running in graph mode, we cannot use BERT pre-trained,
# see https://github.com/huggingface/transformers/issues/3086
# but this is apparently not useful anyway to disable eager mode here, because
# the Keras API compiles models before running them
# from tensorflow.python.framework.ops import disable_eager_execution
# disable_eager_execution()

import datetime

import transformers
from sklearn.metrics import accuracy_score, f1_score, log_loss, precision_recall_fscore_support, r2_score, roc_auc_score
from sklearn.model_selection import train_test_split

from delft.textClassification.config import ModelConfig, TrainingConfig
from delft.textClassification.data_generator import DataGenerator
from delft.textClassification.models import getModel, predict_folds, train_folds
from delft.utilities.Embeddings import Embeddings, load_resource_registry
from delft.utilities.Transformer import DEFAULT_TRANSFORMER_TOKENIZER_DIR, TRANSFORMER_CONFIG_FILE_NAME

transformers.logging.set_verbosity(transformers.logging.ERROR)


class Classifier(object):
    config_file = "config.json"
    weight_file = "model_weights.hdf5"

    def __init__(
        self,
        model_name=None,
        architecture="gru",
        embeddings_name=None,
        list_classes=[],
        char_emb_size=25,
        dropout=0.5,
        recurrent_dropout=0.25,
        use_char_feature=False,
        batch_size=256,
        optimizer="adam",
        learning_rate=None,
        lr_decay=0.9,
        clip_gradients=5.0,
        max_epoch=50,
        patience=5,
        log_dir=None,
        maxlen=300,
        fold_number=1,
        use_roc_auc=True,
        early_stop=True,
        class_weights=None,
        multiprocessing=True,
        num_workers=1,
        transformer_name: str = None,
    ):

        if model_name is None:
            # add a dummy name based on the architecture
            model_name = architecture
            if embeddings_name is not None:
                model_name += "_" + embeddings_name
            if transformer_name is not None:
                model_name += "_" + transformer_name

        if learning_rate is None:
            if transformer_name is None:
                learning_rate = 0.001
            else:
                learning_rate = 2e-5

        self.model = None
        self.models = None
        self.log_dir = log_dir
        self.embeddings_name = embeddings_name
        self.embeddings = None

        # if transformer_name is None, no bert layer is present in the model
        self.transformer_name = None

        self.registry = load_resource_registry("delft/resources-registry.json")

        word_emb_size = 0
        if transformer_name is not None:
            self.transformer_name = transformer_name
            self.embeddings_name = None
            self.embeddings = None
        elif self.embeddings_name is not None:
            self.embeddings = Embeddings(self.embeddings_name, resource_registry=self.registry)
            word_emb_size = self.embeddings.embed_size

        self.model_config = ModelConfig(
            model_name=model_name,
            architecture=architecture,
            embeddings_name=embeddings_name,
            list_classes=list_classes,
            char_emb_size=char_emb_size,
            word_emb_size=word_emb_size,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            use_char_feature=use_char_feature,
            maxlen=maxlen,
            fold_number=fold_number,
            batch_size=batch_size,
            transformer_name=self.transformer_name,
        )

        self.training_config = TrainingConfig(
            learning_rate,
            batch_size=batch_size,
            optimizer=optimizer,
            lr_decay=lr_decay,
            clip_gradients=clip_gradients,
            max_epoch=max_epoch,
            patience=patience,
            use_roc_auc=use_roc_auc,
            early_stop=early_stop,
            class_weights=class_weights,
            multiprocessing=multiprocessing,
            num_workers=num_workers,
        )

    def train(self, x_train, y_train, vocab_init=None, incremental=False, callbacks=None):

        if incremental:
            if self.model is None and self.models is None:
                print("error: you must load a model first for an incremental training")
                return
            print("Incremental training from loaded model", self.model_config.model_name)
        else:
            self.model = getModel(self.model_config, self.training_config)

        print_parameters(self.model_config, self.training_config)
        self.model.print_summary()

        bert_data = False
        if self.transformer_name is not None:
            bert_data = True

        if self.training_config.early_stop:
            # create validation set
            xtr, val_x, y, val_y = train_test_split(x_train, y_train, test_size=0.1)

            training_generator = DataGenerator(
                xtr,
                y,
                batch_size=self.training_config.batch_size,
                maxlen=self.model_config.maxlen,
                list_classes=self.model_config.list_classes,
                embeddings=self.embeddings,
                shuffle=True,
                bert_data=bert_data,
                transformer_tokenizer=self.model.transformer_tokenizer,
            )
            validation_generator = DataGenerator(
                val_x,
                None,
                batch_size=self.training_config.batch_size,
                maxlen=self.model_config.maxlen,
                list_classes=self.model_config.list_classes,
                embeddings=self.embeddings,
                shuffle=False,
                bert_data=bert_data,
                transformer_tokenizer=self.model.transformer_tokenizer,
            )
        else:
            val_y = y_train

            training_generator = DataGenerator(
                x_train,
                y_train,
                batch_size=self.training_config.batch_size,
                maxlen=self.model_config.maxlen,
                list_classes=self.model_config.list_classes,
                embeddings=self.embeddings,
                shuffle=True,
                bert_data=bert_data,
                transformer_tokenizer=self.model.transformer_tokenizer,
            )
            validation_generator = None

        callbacks_ = callbacks if callbacks else []
        callbacks_.append(LogLearningRateCallback(self.model))

        # uncomment to plot graph
        # plot_model(self.model,
        #    to_file='data/models/textClassification/'+self.model_config.model_name+'_'+self.model_config.architecture+'.png')
        self.model.train_model(
            self.model_config.list_classes,
            self.training_config.batch_size,
            self.training_config.max_epoch,
            self.training_config.use_roc_auc,
            self.training_config.class_weights,
            training_generator,
            validation_generator,
            val_y,
            patience=self.training_config.patience,
            multiprocessing=self.training_config.multiprocessing,
            callbacks=callbacks,
        )

    def train_nfold(self, x_train, y_train, vocab_init=None, incremental=False, callbacks=None):
        if incremental:
            if self.models is None:
                print("error: you must load a model first for an incremental training")
                return
            print("Incremental n-fold training from loaded models", self.model_config.model_name)
            self.models = train_folds(
                x_train,
                y_train,
                self.model_config,
                self.training_config,
                self.embeddings,
                self.models,
                callbacks=callbacks,
            )
        else:
            self.models = train_folds(
                x_train, y_train, self.model_config, self.training_config, self.embeddings, None, callbacks=callbacks
            )

    def predict(self, texts, output_format="json", use_main_thread_only=False, batch_size=None):
        bert_data = False
        if self.transformer_name is not None:
            bert_data = True

        if batch_size is not None:
            self.model_config.batch_size = batch_size
            print("---")
            print("batch_size (prediction):", self.model_config.batch_size)
            print("---")

        if self.model_config.fold_number == 1:
            if self.model is not None:
                predict_generator = DataGenerator(
                    texts,
                    None,
                    batch_size=self.model_config.batch_size,
                    maxlen=self.model_config.maxlen,
                    list_classes=self.model_config.list_classes,
                    embeddings=self.embeddings,
                    shuffle=False,
                    bert_data=bert_data,
                    transformer_tokenizer=self.model.transformer_tokenizer,
                )

                result = self.model.predict(predict_generator, use_main_thread_only=use_main_thread_only)
            else:
                raise (OSError("Could not find a model."))
        else:
            if self.models is not None:
                # just a warning: n classifiers using BERT layer for prediction might be heavy in term of model sizes
                predict_generator = DataGenerator(
                    texts,
                    None,
                    batch_size=self.model_config.batch_size,
                    maxlen=self.model_config.maxlen,
                    list_classes=self.model_config.list_classes,
                    embeddings=self.embeddings,
                    shuffle=False,
                    bert_data=bert_data,
                    transformer_tokenizer=self.model.transformer_tokenizer,
                )

                result = predict_folds(
                    self.models,
                    predict_generator,
                    self.model_config,
                    self.training_config,
                    use_main_thread_only=use_main_thread_only,
                )
            else:
                raise (OSError("Could not find nfolds models."))
        if output_format == "json":
            res = {
                "software": "DeLFT",
                "date": datetime.datetime.now().isoformat(),
                "model": self.model_config.model_name,
                "classifications": [],
            }
            i = 0
            for text in texts:
                classification = {"text": text}
                the_res = result[i]
                j = 0
                for cl in self.model_config.list_classes:
                    classification[cl] = float(the_res[j])
                    j += 1
                res["classifications"].append(classification)
                i += 1
            return res
        else:
            return result

    def eval(self, x_test, y_test, use_main_thread_only=False):
        print_parameters(self.model_config, self.training_config)

        bert_data = False
        if self.transformer_name is not None:
            bert_data = True

        if self.model_config.fold_number == 1:
            if self.model is not None:
                self.model.print_summary()
                test_generator = DataGenerator(
                    x_test,
                    None,
                    batch_size=self.model_config.batch_size,
                    maxlen=self.model_config.maxlen,
                    list_classes=self.model_config.list_classes,
                    embeddings=self.embeddings,
                    shuffle=False,
                    bert_data=bert_data,
                    transformer_tokenizer=self.model.transformer_tokenizer,
                )

                result = self.model.predict(test_generator, use_main_thread_only=use_main_thread_only)
            else:
                raise (OSError("Could not find a model."))
        else:
            if self.models is None:
                raise (OSError("Could not find nfolds models."))

            self.models[0].print_summary()

            # just a warning: n classifiers using BERT layer for prediction might be heavy in term of model sizes
            test_generator = DataGenerator(
                x_test,
                None,
                batch_size=self.model_config.batch_size,
                maxlen=self.model_config.maxlen,
                list_classes=self.model_config.list_classes,
                embeddings=self.embeddings,
                shuffle=False,
                bert_data=bert_data,
                transformer_tokenizer=self.models[0].transformer_tokenizer,
            )
            result = predict_folds(
                self.models,
                test_generator,
                self.model_config,
                self.training_config,
                use_main_thread_only=use_main_thread_only,
            )

        print("-----------------------------------------------")
        print("\nEvaluation on", x_test.shape[0], "instances:")

        total_accuracy = 0.0
        total_f1 = 0.0
        total_loss = 0.0
        total_roc_auc = 0.0

        """
        def normer(t):
            if t < 0.5:
                return 0
            else:
                return 1
        vfunc = np.vectorize(normer)
        result_binary = vfunc(result)
        """
        result_intermediate = np.asarray([np.argmax(line) for line in result])

        def vectorize(index, size):
            result = np.zeros(size)
            if index < size:
                result[index] = 1
            return result

        result_binary = np.array([vectorize(xi, len(self.model_config.list_classes)) for xi in result_intermediate])

        precision, recall, fscore, support = precision_recall_fscore_support(y_test, result_binary, average=None)
        print("{:>14}  {:>12}  {:>12}  {:>12}  {:>12}".format(" ", "precision", "recall", "f-score", "support"))
        p = 0
        for the_class in self.model_config.list_classes:
            the_class = the_class[:14]
            print(
                "{:>14}  {:>12}  {:>12}  {:>12}  {:>12}".format(
                    the_class,
                    "{:10.4f}".format(precision[p]),
                    "{:10.4f}".format(recall[p]),
                    "{:10.4f}".format(fscore[p]),
                    support[p],
                )
            )
            p += 1

        # macro-average (average of class scores)
        # we distinguish 1-class and multiclass problems
        if len(self.model_config.list_classes) == 1:
            total_accuracy = accuracy_score(y_test, result_binary)
            total_f1 = f1_score(y_test, result_binary)

            # sklearn will complain if log(0)
            total_loss = log_loss(y_test, result, labels=[0, 1])
            if len(np.unique(y_test)) == 1:
                # roc_auc_score sklearn implementation is not working in this case, it needs more balanced batches
                # a simple fix is to return the r2_score instead in this case (which is a regression score and not a loss)
                total_roc_auc = r2_score(y_test, result)
                if total_roc_auc < 0:
                    total_roc_auc = 0
            else:
                total_roc_auc = roc_auc_score(y_test, result)
        else:
            for j in range(0, len(self.model_config.list_classes)):
                accuracy = accuracy_score(y_test[:, j], result_binary[:, j])
                total_accuracy += accuracy
                f1 = f1_score(y_test[:, j], result_binary[:, j], average="micro")
                total_f1 += f1
                loss = log_loss(y_test[:, j], result[:, j], labels=[0, 1])
                total_loss += loss
                if len(np.unique(y_test[:, j])) == 1:
                    # roc_auc_score sklearn implementation is not working in this case, it needs more balanced batches
                    # a simple fix is to return the r2_score instead in this case (which is a regression score and not a loss)
                    roc_auc = r2_score(y_test[:, j], result[:, j])
                    if roc_auc < 0:
                        roc_auc = 0
                else:
                    roc_auc = roc_auc_score(y_test[:, j], result[:, j], labels=[0, 1])
                total_roc_auc += roc_auc
                """
                print("\nClass:", self.model_config.list_classes[j])
                print("\taccuracy at 0.5 =", accuracy)
                print("\tf-1 at 0.5 =", f1)
                print("\tlog-loss =", loss)
                print("\troc auc =", roc_auc)
                """

        total_accuracy /= len(self.model_config.list_classes)
        total_f1 /= len(self.model_config.list_classes)
        total_loss /= len(self.model_config.list_classes)
        total_roc_auc /= len(self.model_config.list_classes)

        """
        if len(self.model_config.list_classes) != 1:
            print("\nMacro-average:")
        print("\taverage accuracy at 0.5 =", "{:10.4f}".format(total_accuracy))
        print("\taverage f-1 at 0.5 =", "{:10.4f}".format(total_f1))
        print("\taverage log-loss =","{:10.4f}".format( total_loss))
        print("\taverage roc auc =", "{:10.4f}".format(total_roc_auc))
        """

        # micro-average (average of scores for each instance)
        # make sense only if we have more than 1 class, otherwise same as
        # macro-avergae
        if len(self.model_config.list_classes) != 1:
            total_accuracy = 0.0
            total_f1 = 0.0
            total_loss = 0.0
            total_roc_auc = 0.0

            for i in range(0, result.shape[0]):
                accuracy = accuracy_score(y_test[i, :], result_binary[i, :])
                total_accuracy += accuracy
                f1 = f1_score(y_test[i, :], result_binary[i, :], average="micro")
                total_f1 += f1
                loss = log_loss(y_test[i, :], result[i, :], labels=[0.0, 1.0])
                total_loss += loss
                if len(np.unique(y_test[i, :])) == 1:
                    # roc_auc_score sklearn implementation is not working in this case, it needs more balanced batches
                    # a simple fix is to return the r2_score instead in this case (which is a regression score and not a loss)
                    roc_auc = r2_score(y_test[i, :], result[i, :])
                    if roc_auc < 0:
                        roc_auc = 0
                else:
                    roc_auc = roc_auc_score(y_test[i, :], result[i, :], labels=[0.0, 1.0])
                total_roc_auc += roc_auc

            total_accuracy /= result.shape[0]
            total_f1 /= result.shape[0]
            total_loss /= result.shape[0]
            total_roc_auc /= result.shape[0]

            """
            print("\nMicro-average:")
            print("\taverage accuracy at 0.5 =", "{:10.4f}".format(total_accuracy))
            print("\taverage f-1 at 0.5 =", "{:10.4f}".format(total_f1))
            print("\taverage log-loss =", "{:10.4f}".format(total_loss))
            print("\taverage roc auc =", "{:10.4f}".format(total_roc_auc))
            """

    def save(self, dir_path="data/models/textClassification/"):
        # create subfolder for the model if not already exists
        directory = os.path.join(dir_path, self.model_config.model_name)
        if not os.path.exists(directory):
            os.makedirs(directory)

        self.model_config.save(os.path.join(directory, self.config_file))
        print("model config file saved")

        if self.model_config.fold_number == 1:
            if self.model is not None:
                self.model.save(os.path.join(directory, self.weight_file))
                print("model saved")
            else:
                print("Error: model has not been built")
        else:
            if self.models is None:
                print("Error: nfolds models have not been built")
            else:
                # fold models having a transformer layers are already saved
                if self.model_config.transformer_name is None:
                    for i in range(0, self.model_config.fold_number):
                        self.models[i].save(os.path.join(directory, "model{0}_weights.hdf5".format(i)))
                    print("nfolds model saved")

        # save pretrained transformer config and tokenizer if used in the model and if single fold (otherwise it is saved in the nfold process)
        if self.transformer_name is not None and self.model_config.fold_number == 1:
            if self.model.transformer_config is not None:
                self.model.transformer_config.to_json_file(os.path.join(directory, TRANSFORMER_CONFIG_FILE_NAME))
            if self.model.transformer_tokenizer is not None:
                self.model.transformer_tokenizer.save_pretrained(
                    os.path.join(directory, DEFAULT_TRANSFORMER_TOKENIZER_DIR)
                )

    def load(self, dir_path="data/models/textClassification/"):
        model_path = os.path.join(dir_path, self.model_config.model_name)
        self.model_config = ModelConfig.load(os.path.join(model_path, self.config_file))

        if self.model_config.transformer_name is None:
            # load embeddings
            # Do not use cache in 'production' mode
            self.embeddings = Embeddings(
                self.model_config.embeddings_name, resource_registry=self.registry, use_cache=False
            )
            self.model_config.word_embedding_size = self.embeddings.embed_size
        else:
            self.transformer_name = self.model_config.transformer_name
            self.embeddings = None

        self.model = getModel(
            self.model_config, self.training_config, load_pretrained_weights=False, local_path=model_path
        )
        print_parameters(self.model_config, self.training_config)
        self.model.print_summary()

        if self.model_config.fold_number == 1:
            print("load weights from", os.path.join(model_path, self.weight_file))
            self.model.load(os.path.join(model_path, self.weight_file))
        else:
            self.models = []
            if self.model_config.transformer_name is None:
                for i in range(0, self.model_config.fold_number):
                    local_model = getModel(
                        self.model_config, self.training_config, load_pretrained_weights=False, local_path=model_path
                    )
                    local_model.load(os.path.join(model_path, "model{0}_weights.hdf5".format(i)))
                    self.models.append(local_model)
            else:
                # only init first fold one, the other will be init at prediction time, all weights will be loaded at prediction time
                local_model = getModel(
                    self.model_config, self.training_config, load_pretrained_weights=False, local_path=model_path
                )
                self.models.append(local_model)



import os

import numpy as np
import tensorflow as tf
import tf_keras
from tf_keras.callbacks import Callback, EarlyStopping, ModelCheckpoint
from transformers import create_optimizer

from delft.sequenceLabelling.config import ModelConfig
from delft.sequenceLabelling.data_generator import DataGeneratorTransformers
from delft.sequenceLabelling.evaluation import (
    accuracy_score,
    compute_metrics,
    f1_score,
    get_report,
    precision_score,
    recall_score,
)
from delft.sequenceLabelling.models import get_model
from delft.sequenceLabelling.preprocess import Preprocessor
from delft.utilities.misc import print_parameters
from delft.utilities.multiprocessing import get_multiprocessing_config
from delft.utilities.Transformer import DEFAULT_TRANSFORMER_TOKENIZER_DIR, TRANSFORMER_CONFIG_FILE_NAME

DEFAULT_WEIGHT_FILE_NAME = "model_weights.hdf5"
CONFIG_FILE_NAME = "config.json"
PROCESSOR_FILE_NAME = "preprocessor.json"


class Trainer(object):
    def __init__(
        self,
        model,
        models,
        embeddings,
        model_config: ModelConfig,
        training_config,
        checkpoint_path="",
        save_path="",
        preprocessor: Preprocessor = None,
        transformer_preprocessor=None,
        enable_wandb=False,
        nb_workers=6,
    ):

        # for single model training
        self.model = model

        # for n-folds training
        self.models = models

        self.nb_workers = nb_workers

        self.embeddings = embeddings
        self.model_config = model_config
        self.training_config = training_config
        self.checkpoint_path = checkpoint_path
        self.save_path = save_path
        self.preprocessor = preprocessor
        self.transformer_preprocessor = transformer_preprocessor
        self.enable_wandb = enable_wandb

    def train(
        self,
        x_train,
        y_train,
        x_valid,
        y_valid,
        features_train: np.array = None,
        features_valid: np.array = None,
        callbacks=None,
    ):
        """
        Train the instance self.model
        """
        self.model = self.compile_model(self.model, len(x_train))

        # uncomment to plot graph
        # plot_model(self.model,
        #    to_file='data/models/sequenceLabelling/'+self.model_config.model_name+'_'+self.model_config.architecture+'.png')

        self.model = self.train_model(
            self.model,
            x_train,
            y_train,
            x_valid=x_valid,
            y_valid=y_valid,
            f_train=features_train,
            f_valid=features_valid,
            max_epoch=self.training_config.max_epoch,
            callbacks=callbacks,
        )

    def compile_model(self, local_model, train_size):

        nb_train_steps = (train_size // self.training_config.batch_size) * self.training_config.max_epoch

        if self.model_config.transformer_name is not None:
            # we use a transformer layer in the architecture
            optimizer, lr_schedule = create_optimizer(
                init_lr=self.training_config.learning_rate,
                num_train_steps=nb_train_steps,
                weight_decay_rate=0.01,
                num_warmup_steps=0.1 * nb_train_steps,
            )

            if local_model.config.use_chain_crf:
                local_model.compile(optimizer=optimizer, loss=local_model.crf.sparse_crf_loss_bert_masked)
            elif local_model.config.use_crf:
                # loss is calculated by the custom CRF wrapper
                local_model.compile(
                    optimizer=optimizer,
                )
            else:
                # we apply a mask on the predicted labels so that the weights
                # corresponding to special symbols are neutralized
                local_model.compile(
                    optimizer=optimizer,
                    loss=sparse_crossentropy_masked,
                )
        else:
            lr_schedule = tf_keras.optimizers.schedules.ExponentialDecay(
                initial_learning_rate=self.training_config.learning_rate, decay_steps=nb_train_steps, decay_rate=0.1
            )
            optimizer = tf_keras.optimizers.Adam(learning_rate=lr_schedule)

            # optimizer = tf.keras.optimizers.Adam(self.training_config.learning_rate)
            if local_model.config.use_chain_crf:
                local_model.compile(
                    optimizer=optimizer,
                    loss=local_model.crf.loss,
                )
            elif local_model.config.use_crf:
                if tf.executing_eagerly():
                    # loss is calculated by the custom CRF wrapper, no need to specify a loss function here
                    local_model.compile(
                        optimizer=optimizer,
                    )
                else:
                    print("compile model, graph mode")
                    # always expecting a loss function here, but it is calculated internally by the CRF wapper
                    # the following will fail in graph mode because
                    # '<tf.Variable 'chain_kernel:0' shape=(10, 10) dtype=float32> has `None` for gradient.'
                    # however this variable cannot be accessed, so no soluton for the moment
                    # (probably need not using keras fit and creating a custom training loop to get the gradient)
                    local_model.compile(
                        optimizer=optimizer,
                        loss="sparse_categorical_crossentropy",
                    )
                    # local_model.compile(optimizer=optimizer, loss=InnerLossPusher(local_model))
            else:
                # only sparse label encoding is used (no one-hot encoded labels as it was the case in DeLFT < 0.3)
                local_model.compile(
                    optimizer=optimizer,
                    loss="sparse_categorical_crossentropy",
                )

        return local_model

    def train_model(
        self,
        local_model,
        x_train,
        y_train,
        f_train=None,
        x_valid=None,
        y_valid=None,
        f_valid=None,
        max_epoch=50,
        callbacks=None,
    ):
        """
        The parameter model local_model must be compiled before calling this method.
        This model will be returned with trained weights
        """
        # todo: if valid set is None, create it as random segment of the shuffled train set

        generator = local_model.get_generator()
        if self.training_config.early_stop:
            training_generator = generator(
                x_train,
                y_train,
                batch_size=self.training_config.batch_size,
                preprocessor=self.preprocessor,
                bert_preprocessor=self.transformer_preprocessor,
                char_embed_size=self.model_config.char_embedding_size,
                max_sequence_length=self.model_config.max_sequence_length,
                embeddings=self.embeddings,
                shuffle=True,
                features=f_train,
                use_chain_crf=self.model_config.use_chain_crf,
            )

            validation_generator = generator(
                x_valid,
                y_valid,
                batch_size=self.training_config.batch_size,
                preprocessor=self.preprocessor,
                bert_preprocessor=self.transformer_preprocessor,
                char_embed_size=self.model_config.char_embedding_size,
                max_sequence_length=self.model_config.max_sequence_length,
                embeddings=self.embeddings,
                shuffle=False,
                features=f_valid,
                output_input_offsets=True,
                use_chain_crf=self.model_config.use_chain_crf,
            )

            _callbacks = get_callbacks(
                log_dir=self.checkpoint_path,
                early_stopping=True,
                patience=self.training_config.patience,
                valid=(validation_generator, self.preprocessor),
                use_crf=self.model_config.use_crf,
                use_chain_crf=self.model_config.use_chain_crf,
                model=local_model,
                external_callbacks=callbacks,
            )
        else:
            x_train = np.concatenate((x_train, x_valid), axis=0)
            y_train = np.concatenate((y_train, y_valid), axis=0)
            feature_all = None
            if f_train is not None:
                feature_all = np.concatenate((f_train, f_valid), axis=0)

            training_generator = generator(
                x_train,
                y_train,
                batch_size=self.training_config.batch_size,
                preprocessor=self.preprocessor,
                bert_preprocessor=self.transformer_preprocessor,
                char_embed_size=self.model_config.char_embedding_size,
                max_sequence_length=self.model_config.max_sequence_length,
                embeddings=self.embeddings,
                shuffle=True,
                features=feature_all,
                use_chain_crf=self.model_config.use_chain_crf,
            )

            _callbacks = get_callbacks(
                log_dir=self.checkpoint_path,
                early_stopping=False,
                use_crf=self.model_config.use_crf,
                use_chain_crf=self.model_config.use_chain_crf,
                model=local_model,
                external_callbacks=callbacks,
            )
        nb_workers, multiprocessing = get_multiprocessing_config(self.training_config, self.model_config)

        local_model.fit(
            training_generator,
            epochs=max_epoch,
            use_multiprocessing=multiprocessing,
            workers=nb_workers,
            callbacks=_callbacks,
        )

        return local_model

    def train_nfold(self, x_train, y_train, x_valid=None, y_valid=None, f_train=None, f_valid=None, callbacks=None):
        """
        n-fold training for the instance model

        for RNN models:
        -> the n models are stored in self.models, and self.model left unset at this stage
        fold number is available with self.model_config.fold_number

        for models with transformer layer:
        -> fold models are saved on disk (because too large) and self.models is not used, we identify the usage
        of folds with self.model_config.fold_number
        """

        fold_count = self.model_config.fold_number
        fold_size = len(x_train) // fold_count

        dir_path = "data/models/sequenceLabelling/"
        output_directory = os.path.join(dir_path, self.model_config.model_name)
        print("Output directory:", output_directory)
        if not os.path.exists(output_directory):
            os.makedirs(output_directory)

        if self.model_config.transformer_name is not None:
            # save the config, preprocessor and transformer layer config on disk
            self.model_config.save(os.path.join(output_directory, CONFIG_FILE_NAME))
            self.preprocessor.save(os.path.join(output_directory, PROCESSOR_FILE_NAME))

        for fold_id in range(0, fold_count):
            if x_valid is None:
                # segment train and valid
                fold_start = fold_size * fold_id
                fold_end = fold_start + fold_size

                if fold_id == fold_size - 1:
                    fold_end = len(x_train)

                train_x = np.concatenate([x_train[:fold_start], x_train[fold_end:]])
                train_y = np.concatenate([y_train[:fold_start], y_train[fold_end:]])
                train_f = np.concatenate([f_train[:fold_start], f_train[fold_end:]])

                val_x = x_train[fold_start:fold_end]
                val_y = y_train[fold_start:fold_end]
                val_f = f_train[fold_start:fold_end]
            else:
                # reuse given segmentation
                train_x = x_train
                train_y = y_train
                train_f = f_train

                val_x = x_valid
                val_y = y_valid
                val_f = f_valid

            foldModel = get_model(
                self.model_config,
                self.preprocessor,
                ntags=len(self.preprocessor.vocab_tag),
                load_pretrained_weights=True,
            )

            if fold_id == 0:
                print_parameters(self.model_config, self.training_config)
                foldModel.print_summary()

            print("\n------------------------ fold " + str(fold_id) + "--------------------------------------")
            self.transformer_preprocessor = foldModel.transformer_preprocessor
            foldModel = self.compile_model(foldModel, len(train_x))
            foldModel = self.train_model(
                foldModel,
                train_x,
                train_y,
                x_valid=val_x,
                y_valid=val_y,
                f_train=train_f,
                f_valid=val_f,
                max_epoch=self.training_config.max_epoch,
                callbacks=callbacks,
            )

            if self.model_config.transformer_name is None:
                self.models.append(foldModel)
            else:
                # save the model with transformer layer on disk
                weight_file = DEFAULT_WEIGHT_FILE_NAME.replace(".hdf5", str(fold_id) + ".hdf5")
                foldModel.save(os.path.join(output_directory, weight_file))
                if fold_id == 0:
                    foldModel.transformer_config.to_json_file(
                        os.path.join(output_directory, TRANSFORMER_CONFIG_FILE_NAME)
                    )
                    if self.model_config.transformer_name is not None:
                        transformer_preprocessor = foldModel.transformer_preprocessor
                        transformer_preprocessor.tokenizer.save_pretrained(
                            os.path.join(output_directory, DEFAULT_TRANSFORMER_TOKENIZER_DIR)
                        )


class LogLearningRateCallback(Callback):
    def __init__(self, model=None):
        super().__init__()
        self.set_model(model)

    def on_epoch_end(self, epoch, logs):
        if self.model is not None:
            opt = self.model.optimizer
            lr = opt.learning_rate
            # If lr is a schedule, call it with the current step
            if callable(lr):
                lr = lr(opt.iterations)
            # If it's a tf.Variable, get its value
            elif hasattr(lr, "numpy"):
                lr = lr.numpy()
            logs.update({"lr": lr})


def get_callbacks(
    log_dir=None,
    valid=(),
    early_stopping=True,
    patience=5,
    use_crf=True,
    use_chain_crf=False,
    model=None,
    external_callbacks=None,
):
    """
    Get callbacks.

    Args:
        log_dir (str): the destination to save logs
        valid (tuple): data for validation.
        early_stopping (bool): whether to use early stopping.

    Returns:
        list: list of callbacks
    """
    callbacks = []

    if valid:
        callbacks.append(Scorer(*valid, use_crf=use_crf, use_chain_crf=use_chain_crf))

    if log_dir:
        if not os.path.exists(log_dir):
            print("Successfully made a directory: {}".format(log_dir))
            os.mkdir(log_dir)

        file_name = "_".join(["model_weights", "{epoch:02d}"]) + ".h5"
        save_callback = ModelCheckpoint(os.path.join(log_dir, file_name), monitor="f1", save_weights_only=True)
        callbacks.append(save_callback)

    if early_stopping:
        callbacks.append(EarlyStopping(monitor="f1", patience=patience, mode="max"))

    callbacks.append(LogLearningRateCallback(model))

    if external_callbacks:
        callbacks.extend(external_callbacks)

    return callbacks


class Scorer(Callback):
    def __init__(self, validation_generator, preprocessor=None, evaluation=False, use_crf=False, use_chain_crf=False):
        """
        If evaluation is True, we produce a full evaluation with complete report, otherwise it is a
        validation step and will simply produce f1 score
        """
        super(Scorer, self).__init__()
        self.valid_steps = len(validation_generator)
        self.valid_batches = validation_generator
        self.p = preprocessor

        self.f1 = -1.0
        self.accuracy = -1.0
        self.precision = -1.0
        self.recall = -1.0
        self.report = None
        self.report_as_map = None
        self.evaluation = evaluation
        self.use_crf = use_crf
        self.use_chain_crf = use_chain_crf

    def on_epoch_end(self, epoch, logs={}):
        y_pred = None
        y_true = None
        for i, (data, label) in enumerate(self.valid_batches):
            if i == self.valid_steps:
                break
            y_true_batch = label

            if isinstance(self.valid_batches, DataGeneratorTransformers):
                y_true_batch = np.asarray(y_true_batch, dtype=object)

                # we need to remove one vector of the data corresponding to the token offsets, this vector is not
                # expected by the model, but we need it to restore correctly the labels (which are produced
                # according to the sub-segmentation of wordpiece, not the expected segmentation)
                input_offsets = data[-1]
                data = data[:-1]

                y_pred_batch = self.model.predict_on_batch(data)

                if not self.use_crf:
                    y_pred_batch = np.argmax(y_pred_batch, -1)

                if self.use_chain_crf:
                    y_pred_batch = np.argmax(y_pred_batch, -1)

                # results have been produced by a model using a transformer layer, so a few things to do
                # the labels are sparse, so integers and not one hot encoded
                # we need to restore back the labels for wordpiece to the labels for normal tokens
                # for this we can use the marked tokens provided by the generator
                new_y_pred_batch = []
                new_y_true_batch = []
                for y_pred_text, y_true_text, offsets_text in zip(y_pred_batch, y_true_batch, input_offsets):
                    new_y_pred_text = []
                    new_y_true_text = []
                    # this is the result per sequence, realign labels:
                    for q in range(len(offsets_text)):
                        if offsets_text[q][0] == 0 and offsets_text[q][1] == 0:
                            # special token
                            continue
                        if offsets_text[q][0] != 0:
                            # added sub-token
                            continue
                        new_y_pred_text.append(y_pred_text[q])
                        new_y_true_text.append(y_true_text[q])
                    new_y_pred_batch.append(new_y_pred_text)
                    new_y_true_batch.append(new_y_true_text)
                y_pred_batch = new_y_pred_batch
                y_true_batch = new_y_true_batch

                y_true_batch = [self.p.inverse_transform(y) for y in y_true_batch]
                y_pred_batch = [self.p.inverse_transform(y) for y in y_pred_batch]
            else:
                # no transformer layer around, no mess to manage with the sub-tokenization...
                y_pred_batch = self.model.predict_on_batch(data)

                if not self.use_crf:
                    # one hot encoded predictions
                    y_pred_batch = np.argmax(y_pred_batch, -1)

                if self.use_chain_crf:
                    # one hot encoded predictions and labels
                    y_pred_batch = np.argmax(y_pred_batch, -1)
                    y_true_batch = np.argmax(y_true_batch, -1)

                # we also have the input length available
                sequence_lengths = data[-1]  # this is the vectors "length_input" of the models input, always last
                # shape of (batch_size, 1), we want (batch_size)
                sequence_lengths = np.reshape(sequence_lengths, (-1,))

                y_pred_batch = [self.p.inverse_transform(y[:l]) for y, l in zip(y_pred_batch, sequence_lengths)]
                y_true_batch = [self.p.inverse_transform(y[:l]) for y, l in zip(y_true_batch, sequence_lengths)]

            if i == 0:
                y_pred = y_pred_batch
                y_true = y_true_batch
            else:
                y_pred.extend(y_pred_batch)
                y_true.extend(y_true_batch)

        """
        for i in range(0,len(y_pred)):
            print("pred", y_pred[i])
            print("true", y_true[i])
        """
        has_data = y_true is not None and y_pred is not None
        f1 = f1_score(y_true, y_pred) if has_data else 0.0
        print("\tf1 (micro): {:04.2f}".format(f1 * 100))

        if self.evaluation:
            self.accuracy = accuracy_score(y_true, y_pred) if has_data else 0.0
            self.precision = precision_score(y_true, y_pred) if has_data else 0.0
            self.recall = recall_score(y_true, y_pred) if has_data else 0.0
            self.report_as_map = compute_metrics(y_true, y_pred) if has_data else compute_metrics([], [])
            self.report = get_report(self.report_as_map, digits=4)
            print(self.report)

        # save eval
        logs["f1"] = f1
        self.f1 = f1


def sparse_crossentropy_masked(y_true, y_pred):
    mask_value = 0
    y_true_masked = tf.boolean_mask(y_true, tf.not_equal(y_true, mask_value))
    y_pred_masked = tf.boolean_mask(y_pred, tf.not_equal(y_true, mask_value))
    return tf.reduce_mean(tf_keras.losses.sparse_categorical_crossentropy(y_true_masked, y_pred_masked))



import tensorflow as tf
import tf_keras
from tensorflow_addons.text.crf_wrapper import CRFModelWrapper

"""
A slightly modified TensorFlow addons CRF wrapper trying to return by default usable scores
with decode_sequence

-> it turns out not working as expected when exploring https://stats.stackexchange.com/a/517325
we don't get probabilities and crf_log_norm normalization is far too low
-> issue on tensorflow addons github.com/tensorflow/addons/issues/2088
-> we will probably need to wait for PR https://github.com/tensorflow/addons/pull/1935 to be merged

"""


@tf_keras.utils.register_keras_serializable(package="Addons")
class CRFModelWrapperDefault(CRFModelWrapper):
    def call(self, inputs, training=None, mask=None, return_crf_internal=False):
        base_model_outputs = self.base_model(inputs, training, mask)

        # change next line, if your model has more outputs
        crf_input = base_model_outputs

        decode_sequence, potentials, sequence_length, kernel = self.crf_layer(crf_input)

        # change next line, if your base model has more outputs
        # Aways keep `(potentials, sequence_length, kernel), decode_sequence, `
        # as first two outputs of model.
        # current `self.train_step()` expected such settings
        if not tf.executing_eagerly():
            decode_sequence = tf.cast(decode_sequence, tf.float32)
        outputs = (potentials, sequence_length, kernel), decode_sequence

        """
        tf.print(potentials)
        tf.print(sequence_length)
        tf.print(kernel)

        # See https://stats.stackexchange.com/a/517325

        # compute the normalization log Z(x) of the CRF from input, lenghts and
        # transition matrix (chain kernel)
        normalization = crf_log_norm(potentials, sequence_length, kernel)
        #sequence_length_float32 = tf.cast(sequence_length, tf.float32)
        #normalization = tf.divide(normalization, sequence_length_float32)
        #tf.print(normalization)

        normalization = tf.expand_dims(normalization, axis=1)
        normalization = tf.expand_dims(normalization, axis=1)

        # then compute the difference
        probabilities = tf.math.subtract(potentials, normalization)

        # and finally exponentiate to get probabilities
        probabilities = tf.math.exp(potentials)
        #tf.print(probabilities)
        """

        if return_crf_internal:
            return outputs
        else:
            # outputs[0] is the crf internal, skip it
            output_without_crf_internal = outputs[1:]

            # it is nicer to return a tensor instead of an one tensor list
            if len(output_without_crf_internal) == 1:
                return output_without_crf_internal[0]
            else:
                return output_without_crf_internal


class InnerLossPusher(tf_keras.losses.Loss):
    """
    Experimental...
    When earger mode is disabled, Keras model.compile() requires a loss function.
    The following custom loss function is simply retrieving the inner model loss calculation
    returning it as explicit loss function for Keras fit()
    """

    def __init__(self, model, name="custom_inner_loss_pusher"):
        super().__init__(name=name)
        self.model = model

    def call(self, y_true, y_pred):
        print(self.model.losses)
        return self.model.losses



# Manage pre-trained embeddings
import gzip
import hashlib
import io
import json
import logging
import mmap
import os
import pickle
import shutil
import struct
import sys
import zipfile

import lmdb
import numpy as np
from tqdm import tqdm

logging.basicConfig()
logging.getLogger().setLevel(logging.ERROR)

# for fasttext binary embeddings
fasttext_support = True
try:
    import fastText
except ImportError:
    fasttext_support = False

from delft.utilities.Utilities import download_file

# gensim is used to exploit .bin FastText embeddings, in particular the OOV with the provided ngrams
# from gensim.models import FastText

# this is the default init size of a lmdb database for embeddings
# based on https://github.com/kermitt2/nerd/blob/master/src/main/java/com/scienceminer/nerd/kb/db/KBDatabase.java
# and https://github.com/kermitt2/nerd/blob/0.0.3/src/main/java/com/scienceminer/nerd/kb/db/KBDatabaseFactory.java#L368
map_size = 100 * 1024 * 1024 * 1024


class Embeddings(object):
    def __init__(self, name, resource_registry=None, lang="en", extension="vec", use_cache=True, load=True):

        self.name = name
        self.embed_size = 0
        self.static_embed_size = 0
        self.vocab_size = 0
        self.model = {}
        self.registry = resource_registry
        self.lang = lang
        self.extension = extension
        self.embedding_lmdb_path = None
        if self.registry is not None and "embedding-lmdb-path" in self.registry:
            self.embedding_lmdb_path = self.registry["embedding-lmdb-path"]
        self.env = None
        if load:
            self.make_embeddings_simple(name)
        self.static_embed_size = self.embed_size

        self.use_cache = use_cache

    def __getattr__(self, name):
        return getattr(self.model, name)

    def __getstate__(self):
        state = self.__dict__.copy()
        state["env"] = None
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        if self.embedding_lmdb_path and os.path.isdir(os.path.join(self.embedding_lmdb_path, self.name)):
            envFilePath = os.path.join(self.embedding_lmdb_path, self.name)
            self.env = lmdb.open(envFilePath, readonly=True, max_readers=2048, max_spare_txns=2, lock=False)
            with self.env.begin() as txn:
                cursor = txn.cursor()
                for key, value in cursor:
                    _check_lmdb_format(value)
                    break
                cursor.close()

    def reopen_lmdb(self):
        """
        Reopen the LMDB environment. This is required for fork-safe multiprocessing.

        LMDB environments opened before fork() cannot be safely used in child processes.
        Call this method in each worker process (e.g., via DataLoader's worker_init_fn)
        to create a fresh LMDB environment handle for the worker.

        This is a no-op if embeddings are not using LMDB (e.g., in-memory or bin format).
        """
        if self.env is None and self.embedding_lmdb_path is None:
            return

        if self.env is not None:
            try:
                self.env.close()
            except Exception:
                pass

        if self.embedding_lmdb_path and os.path.isdir(os.path.join(self.embedding_lmdb_path, self.name)):
            envFilePath = os.path.join(self.embedding_lmdb_path, self.name)
            self.env = lmdb.open(
                envFilePath,
                readonly=True,
                max_readers=2048,
                max_spare_txns=2,
                lock=False,
            )

    def make_embeddings_simple_in_memory(self, name="fasttext-crawl"):
        nbWords = 0
        print("loading embeddings...")
        begin = True
        description = self.get_description(name)
        if description is not None:
            embeddings_path = description["path"]
            self.lang = description["lang"]
            print("path:", embeddings_path)
            if self.extension == "bin":
                self.model = fastText.load_model(embeddings_path)
                nbWords = len(self.model.get_words())
                self.embed_size = self.model.get_dimension()
            else:
                with open(embeddings_path, encoding="utf8") as f:
                    for line in f:
                        line = line.strip().split(" ")
                        if begin:
                            begin = False
                            nb_words, embed_size = _fetch_header_if_available(line)

                            # we parse the header
                            if nb_words > 0 and embed_size > 0:
                                nbWords = nb_words
                                self.embed_size = embed_size
                                continue

                        word = line[0]
                        vector = np.array([float(val) for val in line[1 : len(line)]], dtype="float32")
                        # else:
                        #    vector = np.array([float(val) for val in line[1:len(line)-1]], dtype='float32')
                        if self.embed_size == 0:
                            self.embed_size = len(vector)
                        self.model[word] = vector
                if nbWords == 0:
                    nbWords = len(self.model)
            print("embeddings loaded for", nbWords, "words and", self.embed_size, "dimensions")

    def make_embeddings_lmdb(self, name="fasttext-crawl"):
        print("\nCompiling embeddings... (this is done only one time per embeddings at first usage)")
        description = self.get_description(name)

        if description is None:
            print("Error: embedding name '{}' not found in the resource registry.".format(name))
            raise RuntimeError(
                "Embedding '{}' not found in resources-registry.json. Check the name or add it to the registry.".format(
                    name
                )
            )

        if description is not None:
            # the following method will possibly download the embedding file if not available locally
            embeddings_path = self.get_embedding_path(description)
            if embeddings_path is None:
                print("\nCould not locate a usable resource for embeddings", name)
                return

            self.load_embeddings_from_file(embeddings_path)

        # cleaning possible downloaded embeddings
        self.clean_downloads()

    def load_embeddings_from_file(self, embeddings_path):
        begin = True
        nbWords = 0
        txn = self.env.begin(write=True)
        # batch_size = 1024
        i = 0
        nb_lines = 0

        # read number of lines first
        embedding_file = open_embedding_file(embeddings_path)
        if embedding_file is None:
            print("Error: could not open embeddings file", embeddings_path)
            return

        for line in embedding_file:
            nb_lines += 1
        embedding_file.close()

        embedding_file = open_embedding_file(embeddings_path)
        # with open(embeddings_path, encoding='utf8') as f:
        for line in tqdm(embedding_file, total=nb_lines):
            line = line.decode()
            line = line.split(" ")
            if begin:
                begin = False
                nb_words, embed_size = _fetch_header_if_available(line)

                if nb_words > 0 and embed_size > 0:
                    nbWords = nb_words
                    self.embed_size = embed_size
                    continue

            word = line[0]
            try:
                if line[len(line) - 1] == "\n":
                    vector = np.array([float(val) for val in line[1 : len(line) - 1]], dtype="float32")
                else:
                    vector = np.array([float(val) for val in line[1 : len(line)]], dtype="float32")

                # vector = np.array([float(val) for val in line[1:len(line)]], dtype='float32')
            except Exception:
                print(len(line))
                print(line[1 : len(line)])
            # else:
            #    vector = np.array([float(val) for val in line[1:len(line)-1]], dtype='float32')
            if self.embed_size == 0:
                self.embed_size = len(vector)

            if len(word.encode(encoding="UTF-8")) < self.env.max_key_size():
                txn.put(word.encode(encoding="UTF-8"), _serialize_float32(vector))
                # txn.put(word.encode(encoding='UTF-8'), _serialize_byteio(vector))
                i += 1

            # commit batch
            # if i % batch_size == 0:
            #     txn.commit()
            #     txn = self.env.begin(write=True)

        embedding_file.close()

        # if i % batch_size != 0:
        txn.commit()
        if nbWords == 0:
            nbWords = i
        self.vocab_size = nbWords
        print("embeddings loaded for", nbWords, "words and", self.embed_size, "dimensions")

    def clean_downloads(self):
        # cleaning possible downloaded embeddings
        if (
            "embedding-download-path" in self.registry
            and os.path.exists(self.registry["embedding-download-path"])
            and os.path.isdir(self.registry["embedding-download-path"])
        ):
            for filename in os.listdir(self.registry["embedding-download-path"]):
                file_path = os.path.join(self.registry["embedding-download-path"], filename)
                try:
                    if os.path.isfile(file_path) or os.path.islink(file_path):
                        os.unlink(file_path)
                    elif os.path.isdir(file_path):
                        shutil.rmtree(file_path)
                except Exception as e:
                    print("Failed to delete %s. Reason: %s" % (file_path, e))

    def make_embeddings_simple(self, name="fasttext-crawl"):
        description = self.get_description(name)
        if description is None:
            available = []
            if self.registry is not None:
                for section in ["embeddings", "embeddings-contextualized", "transformers"]:
                    if section in self.registry:
                        available.extend([emb["name"] for emb in self.registry[section]])
            print("Error: embedding name '{}' not found in the resource registry.".format(name))
            if available:
                print("Available embeddings: {}".format(", ".join(available)))
            raise RuntimeError(
                "Embedding '{}' not found in resources-registry.json. Check the name or add it to the registry.".format(
                    name
                )
            )
        if description is not None:
            self.extension = description["format"]

        if self.extension == "bin":
            if fasttext_support:
                print("embeddings are of .bin format, so they will be loaded in memory...")
                self.make_embeddings_simple_in_memory(name)
            else:
                if not (sys.platform == "linux" or sys.platform == "darwin"):
                    raise ValueError("FastText .bin format not supported for your platform")
                else:
                    raise ValueError(
                        "Go to the documentation to get more information on how to install FastText .bin support"
                    )

        elif self.embedding_lmdb_path is None or self.embedding_lmdb_path == "None":
            print(
                "embedding_lmdb_path is not specified in the embeddings registry, so the embeddings will be loaded in memory..."
            )
            embeddings_path = None
            if "path" in description:
                embeddings_path = description["path"]
            self.lang = description["lang"]
            if embeddings_path is None or not os.path.isfile(embeddings_path):
                raise ValueError("Embedding path for", description["name"], "is not valid", embeddings_path)

            self.make_embeddings_simple_in_memory(name)

        else:
            # if the path to the lmdb database files does not exist, we create it
            if not os.path.isdir(self.embedding_lmdb_path):
                # conservative check (likely very useless)
                if not os.path.exists(self.embedding_lmdb_path):
                    os.makedirs(self.embedding_lmdb_path, exist_ok=True)

            # check if the lmdb database exists
            envFilePath = os.path.join(self.embedding_lmdb_path, name)
            load_db = True
            if os.path.isdir(envFilePath):
                description = self.get_description(name)
                if description is not None:
                    self.lang = description["lang"]

                # open the database in read mode
                self.env = lmdb.open(envFilePath, readonly=True, max_readers=2048, max_spare_txns=4)
                if self.env:
                    # we need to set self.embed_size and self.vocab_size
                    with self.env.begin() as txn:
                        stats = txn.stat()
                        size = stats["entries"]
                        self.vocab_size = size

                    with self.env.begin() as txn:
                        cursor = txn.cursor()
                        for key, value in cursor:
                            _check_lmdb_format(value)
                            vector = _deserialize_float32(value)
                            self.embed_size = vector.shape[0]
                            break
                        cursor.close()

                    if self.vocab_size > 100 and self.embed_size > 10:
                        # lmdb database exists and looks valid
                        load_db = False

                        # no idea why, but we need to close and reopen the environment to avoid
                        # mdb_txn_begin: MDB_BAD_RSLOT: Invalid reuse of reader locktable slot
                        # when opening new transaction !
                        self.env.close()
                        self.env = lmdb.open(envFilePath, readonly=True, max_readers=2048, max_spare_txns=2)

            if load_db:
                # create and load the database in write mode
                self.env = lmdb.open(envFilePath, map_size=map_size)
                self.make_embeddings_lmdb(name)

    def get_description(self, name):
        for emb in self.registry["embeddings"]:
            if emb["name"] == name:
                return emb
        for emb in self.registry["embeddings-contextualized"]:
            if emb["name"] == name:
                return emb
        for emb in self.registry["transformers"]:
            if emb["name"] == name:
                return emb
        return None

    def get_word_vector(self, word):
        """
        Get static embeddings (e.g. glove) for a given token
        """
        if (self.name == "wiki.fr") or (self.name == "wiki.fr.bin"):
            # the pre-trained embeddings are not cased
            word = word.lower()
        if self.env is None or self.extension == "bin":
            # db not available or embeddings in bin format, the embeddings should be available in memory (normally!)
            return self.get_word_vector_in_memory(word)
        try:
            with self.env.begin() as txn:
                vector = txn.get(word.encode(encoding="UTF-8"))
                if vector:
                    word_vector = _deserialize_float32(vector)
                    vector = None
                else:
                    word_vector = np.zeros((self.static_embed_size,), dtype=np.float32)
                    # alternatively, initialize with random negative values
                    # word_vector = np.random.uniform(low=-0.5, high=0.0, size=(self.embed_size,))
                    # alternatively use fasttext OOV ngram possibilities (if ngram available)
        except lmdb.Error:
            # no idea why, but we need to close and reopen the environment to avoid
            # mdb_txn_begin: MDB_BAD_RSLOT: Invalid reuse of reader locktable slot
            # when opening new transaction !
            self.env.close()
            envFilePath = os.path.join(self.embedding_lmdb_path, self.name)
            self.env = lmdb.open(envFilePath, readonly=True, max_readers=2048, max_spare_txns=2, lock=False)
            return self.get_word_vector(word)
        return word_vector

    def get_word_vector_in_memory(self, word):
        if (self.name == "wiki.fr") or (self.name == "wiki.fr.bin"):
            # the pre-trained embeddings are not cased
            word = word.lower()
        if self.extension == "bin":
            return self.model.get_word_vector(word)
        if word in self.model:
            return self.model[word]
        else:
            # for unknown word, we use a vector filled with 0.0
            return np.zeros((self.static_embed_size,), dtype=np.float32)
            # alternatively, initialize with random negative values
            # return np.random.uniform(low=-0.5, high=0.0, size=(self.embed_size,))
            # alternatively use fasttext OOV ngram possibilities (if ngram available)

    def get_embedding_path(self, description):
        embeddings_path = None
        if "path" in description:
            embeddings_path = description["path"]
        self.lang = description["lang"]

        if embeddings_path is None or not os.path.isfile(embeddings_path):
            print("error: embedding path for", description["name"], "is not valid", embeddings_path)
            if "url" in description and len(description["url"]) > 0:
                url = description["url"]
                download_path = self.registry["embedding-download-path"]
                # if the download path does not exist, we create it
                if not os.path.isdir(download_path):
                    try:
                        os.mkdir(download_path)
                    except OSError:
                        print("Creation of the download directory", download_path, "failed")

                print("Downloading resource file for", description["name"], "...")
                embeddings_path = download_file(url, download_path)
                if embeddings_path is not None and os.path.isfile(embeddings_path):
                    print("Download sucessful:", embeddings_path)
            else:
                print(
                    "no download url available for this embeddings resource, please review the embedding registry for",
                    description["name"],
                )
        return embeddings_path


def _serialize_byteio(array):
    memfile = io.BytesIO()
    np.save(memfile, array)
    memfile.seek(0)
    return memfile.getvalue()


def _deserialize_byteio(serialized):
    memfile = io.BytesIO()
    memfile.write(serialized)
    memfile.seek(0)
    return np.load(memfile)


def _serialize_pickle(a):
    """Legacy: pickle serialization (for reading old databases)"""
    return pickle.dumps(a)


def _deserialize_pickle(serialized):
    """Legacy: pickle deserialization (for reading old databases)"""
    return pickle.loads(serialized)


def _serialize_float32(array):
    """
    Serialize numpy array to raw float32 bytes.
    This format is readable by both Python and Java.
    """
    return array.astype(np.float32).tobytes()


def _deserialize_float32(serialized):
    """
    Deserialize raw float32 bytes to numpy array.
    This format is readable by both Python and Java.
    """
    return np.frombuffer(serialized, dtype=np.float32)


def _check_lmdb_format(value):
    """
    Check that an LMDB value contains raw float32 bytes, not legacy pickle format.

    Pickle-serialized numpy arrays start with pickle protocol bytes (\\x80 + protocol 2-5)
    and contain b'numpy' in the first 50 bytes. Both signals must be present to avoid
    false positives on raw float32 data that happens to start with \\x80.

    Raises ValueError with conversion instructions if pickle format is detected.
    """
    if len(value) < 2:
        return

    # Signal 1: pickle magic byte (\x80) followed by protocol version 2-5
    if value[0] == 0x80 and value[1] in (2, 3, 4, 5):
        # Signal 2: b'numpy' substring in first 50 bytes (always present in pickled numpy arrays)
        if b"numpy" in value[:50]:
            raise ValueError(
                "LMDB embedding database is in legacy pickle format, which is incompatible "
                "with the current raw float32 format and will produce garbage vectors. "
                "Please convert it using:\n\n"
                "  python -m delft.utilities.convert_lmdb_embeddings "
                "--input <path-to-old-lmdb> --output <path-to-new-lmdb>\n\n"
                "Then update your embedding-lmdb-path in the resource registry to point "
                "to the converted database."
            )


def open_embedding_file(embeddings_path):
    # embeddings can be uncompressed or compressed with gzip or zip
    if embeddings_path.endswith(".gz"):
        embedding_file = gzip.open(embeddings_path, mode="rb")
    elif embeddings_path.endswith(".zip"):
        zip_file = zipfile.ZipFile(embeddings_path, "r")
        for filename in zip_file.namelist():
            print(filename)
            if filename.endswith("vec") or filename.endswith("txt"):
                embedding_file = zip_file.open(filename, mode="r")
    else:
        embedding_file = open(embeddings_path, mode="rb")
    return embedding_file


def _get_num_lines(file_path):
    fp = open(file_path, "r+")
    buf = mmap.mmap(fp.fileno(), 0)
    lines = 0
    while buf.readline():
        lines += 1
    fp.close()
    return lines


def list_digest(strings):
    hash = hashlib.sha1()
    for s in strings:
        hash.update(struct.pack("I", len(s)))
        hash.update(s.encode(encoding="UTF-8"))
    return hash.hexdigest()


def is_int(s):
    try:
        int(s)
        return True
    except ValueError:
        return False


def is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def _fetch_header_if_available(line):
    """
    Fetch the header if the line is just composed by two elements the number of workds and the embedding size

    :param line: a splitted line (if not split, tried to split by spaces)
    :return: the number of words and the embedding size, if there is no header, they will be set to -1
    """
    if isinstance(line, str):
        line = line.split(" ")

    nb_words = -1
    embed_size = -1

    if len(line) == 2:
        # first line gives the nb of words and the embedding size
        nb_words = int(line[0])
        embed_size = int(line[1].replace("\n", ""))

    return nb_words, embed_size


def load_resource_registry(path="delft/resources-registry.json"):
    """
    Load the resource registry file in memory. Each description provides a name,
    a file path (used only if necessary) and an embeddings type (to take into account
    small variation of format)
    """
    registry_json = open(path).read()
    return json.loads(registry_json)



import os
from typing import Union

os.environ["KERAS_BACKEND"] = "tensorflow"

from transformers import AutoConfig, AutoTokenizer, BertTokenizer, TFAutoModel, TFBertModel

TRANSFORMER_CONFIG_FILE_NAME = "transformer-config.json"
DEFAULT_TRANSFORMER_TOKENIZER_DIR = "transformer-tokenizer"

LOADING_METHOD_LOCAL_MODEL_DIR = "local_model_dir"
LOADING_METHOD_HUGGINGFACE_NAME = "huggingface"
LOADING_METHOD_PLAIN_MODEL = "plain_model"
LOADING_METHOD_DELFT_MODEL = "delft_model"


class Transformer(object):
    """
    This class provides a wrapper around a transformer model (pre-trained or fine-tuned)

    This class makes possible to load a transformer config, tokenizer and weights using different approaches,
    prioritizing the most "local" method over an external access to the HuggingFace Hub:

     1. loading the transformer saved locally as part of an existing full delft model (the configuration file
        name will be different)
     2. via a local directory
     3. by specifying the weight, config and vocabulary files separately (this is likely the scenario where
        the user try to load a model from the downloaded bert/scibertt model from github
     4. via the HuggingFace transformer name and HuggingFace Hub, resulting in several online requests to this
        Hub, which could fail if the service is overloaded
    """

    def __init__(self, name: str, resource_registry: dict = None, delft_local_path: str = None):

        self.bert_preprocessor = None
        self.transformer_config = None
        self.loading_method = None
        self.model = None

        # In case the model is loaded from a local directory
        self.local_dir_path = None

        # In case the weights, config and vocab are specified separately (model vanilla)
        self.local_weights_file = None
        self.local_config_file = None
        self.local_vocab_file = None

        self.name = name

        if delft_local_path:
            self.loading_method = LOADING_METHOD_DELFT_MODEL
            self.local_dir_path = delft_local_path

        self.tokenizer = None

        if resource_registry:
            self.configure_from_registry(resource_registry)

        self.auth_token = None

        # read possible Hugging Face access token to support private models
        # will be None if the key does not exist
        self.auth_token = os.getenv("HF_ACCESS_TOKEN")

    def configure_from_registry(self, resource_registry) -> None:
        """
        Fetch transformer information from the registry and infer the loading method:
            1. if no configuration is provided is using HuggingFace with the provided name
            2. if only the directory is provided it will load the model from that directory
            3. if the weights, config and vocab are provided (as in the vanilla models) then
               it will load them as BertTokenizer and BertModel
        """

        if self.loading_method == LOADING_METHOD_DELFT_MODEL:
            return

        if "transformers" in resource_registry:
            filtered_resources = list(
                filter(lambda x: "name" in x and x["name"] == self.name, resource_registry["transformers"])
            )
            if len(filtered_resources) > 0:
                transformer_configuration = list(filtered_resources)[0]
                if "model_dir" in transformer_configuration:
                    self.local_dir_path = transformer_configuration["model_dir"]
                    self.loading_method = LOADING_METHOD_LOCAL_MODEL_DIR
                else:
                    self.loading_method = LOADING_METHOD_PLAIN_MODEL
                    if "path-config" in transformer_configuration and os.path.isfile(
                        transformer_configuration["path-config"]
                    ):
                        self.local_config_file = transformer_configuration["path-config"]
                    else:
                        print("Missing path-config or not a file.")

                    if (
                        "path-weights" in transformer_configuration
                        and os.path.isfile(transformer_configuration["path-weights"])
                        or os.path.isfile(transformer_configuration["path-weights"] + ".data-00000-of-00001")
                    ):
                        self.local_weights_file = transformer_configuration["path-weights"]
                    else:
                        print("Missing weights-config or not a file.")

                    if "path-vocab" in transformer_configuration and os.path.isfile(
                        transformer_configuration["path-vocab"]
                    ):
                        self.local_vocab_file = transformer_configuration["path-vocab"]
                    else:
                        print("Missing vocab-file or not a file.")
            else:
                self.loading_method = LOADING_METHOD_HUGGINGFACE_NAME
                # print(No configuration for", self.name, "Loading from Hugging face.")
        else:
            self.loading_method = LOADING_METHOD_HUGGINGFACE_NAME
            # print("No configuration for", self.name, "Loading from Hugging face.")

    def init_preprocessor(self, max_sequence_length: int, add_prefix_space: bool = True):
        """
        Load the tokenizer according to the provided information, in case of missing configuration,
        it will try to use huggingface as fallback solution.
        """
        if self.loading_method == LOADING_METHOD_HUGGINGFACE_NAME:
            # fix for model without tokenizer config on HuggingFace which might default to
            # invalid casing (e.g. allenai/scibert_scivocab_cased, defaulting to uncase see #144)
            # and when model has the case information explicitly in its name
            do_lower_case = None
            if str.lower(self.name).find("uncased") != -1:
                do_lower_case = True
            elif str.lower(self.name).find("cased") != -1:
                do_lower_case = False

            if do_lower_case is not None:
                if self.auth_token is not None:
                    self.tokenizer = AutoTokenizer.from_pretrained(
                        self.name,
                        max_length=max_sequence_length,
                        add_prefix_space=add_prefix_space,
                        do_lower_case=do_lower_case,
                        token=self.auth_token,
                    )
                else:
                    self.tokenizer = AutoTokenizer.from_pretrained(
                        self.name,
                        max_length=max_sequence_length,
                        add_prefix_space=add_prefix_space,
                        do_lower_case=do_lower_case,
                    )

            else:
                if self.auth_token is not None:
                    self.tokenizer = AutoTokenizer.from_pretrained(
                        self.name,
                        max_length=max_sequence_length,
                        add_prefix_space=add_prefix_space,
                        token=self.auth_token,
                    )
                else:
                    self.tokenizer = AutoTokenizer.from_pretrained(
                        self.name, max_length=max_sequence_length, add_prefix_space=add_prefix_space
                    )

        elif self.loading_method == LOADING_METHOD_LOCAL_MODEL_DIR:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.local_dir_path, max_length=max_sequence_length, add_prefix_space=add_prefix_space
            )
        elif self.loading_method == LOADING_METHOD_PLAIN_MODEL:
            self.tokenizer = BertTokenizer.from_pretrained(self.local_vocab_file)

        elif self.loading_method == LOADING_METHOD_DELFT_MODEL:
            config_path = os.path.join(".", self.local_dir_path, TRANSFORMER_CONFIG_FILE_NAME)
            self.transformer_config = AutoConfig.from_pretrained(config_path)
            self.tokenizer = AutoTokenizer.from_pretrained(
                os.path.join(self.local_dir_path, DEFAULT_TRANSFORMER_TOKENIZER_DIR), config=self.transformer_config
            )

    def save_tokenizer(self, output_directory):
        self.tokenizer.save_pretrained(output_directory)

    def instantiate_layer(self, load_pretrained_weights=True) -> Union[object, TFAutoModel, TFBertModel]:
        """
        Instantiate a transformer to be loaded in a Keras layer using the availability method of the pre-trained transformer.
        """
        if self.loading_method == LOADING_METHOD_HUGGINGFACE_NAME:
            if load_pretrained_weights:
                if self.auth_token is not None:
                    try:
                        transformer_model = TFAutoModel.from_pretrained(self.name, from_pt=True, token=self.auth_token)
                    except Exception:
                        # failure might be due to safetensors format for the weights, we can try an alternative loading
                        # for this case
                        transformer_model = TFAutoModel.from_pretrained(self.name, token=self.auth_token)
                else:
                    try:
                        transformer_model = TFAutoModel.from_pretrained(self.name, from_pt=True)
                    except Exception:
                        # failure might be due to safetensors format for the weights, we can try an alternative loading
                        # for this case
                        transformer_model = TFAutoModel.from_pretrained(self.name)

                self.transformer_config = transformer_model.config
                return transformer_model
            else:
                config_path = os.path.join(".", self.local_dir_path, TRANSFORMER_CONFIG_FILE_NAME)
                self.transformer_config = AutoConfig.from_pretrained(config_path)
                return TFAutoModel.from_config(self.transformer_config)

        elif self.loading_method == LOADING_METHOD_LOCAL_MODEL_DIR:
            if load_pretrained_weights:
                try:
                    transformer_model = TFAutoModel.from_pretrained(self.local_dir_path, from_pt=True)
                except Exception:
                    # failure might be due to safetensors format for the weights, we can try an alternative loading
                    # for this case
                    transformer_model = TFAutoModel.from_pretrained(self.local_dir_path)
                self.transformer_config = transformer_model.config
                return transformer_model
            else:
                config_path = os.path.join(".", self.local_dir_path, TRANSFORMER_CONFIG_FILE_NAME)
                self.transformer_config = AutoConfig.from_pretrained(config_path)
                # self.transformer_config = AutoConfig.from_pretrained(self.local_dir_path)
                return TFAutoModel.from_config(self.transformer_config)

        elif self.loading_method == LOADING_METHOD_PLAIN_MODEL:
            if load_pretrained_weights:
                self.transformer_config = AutoConfig.from_pretrained(self.local_config_file)
                # transformer_model = TFBertModel.from_pretrained(self.local_weight_file, from_tf=True)
                raise NotImplementedError(
                    "The load of TF weights from huggingface automodel classes is not yet implemented. \
                    Please use load from Hugging Face Hub or from directory for the initial loading of the transformers weights."
                )
            else:
                config_path = os.path.join(".", self.local_dir_path, TRANSFORMER_CONFIG_FILE_NAME)
                self.transformer_config = AutoConfig.from_pretrained(config_path)
                return TFBertModel.from_config(self.transformer_config)

        else:
            # TODO: revise this
            if load_pretrained_weights:
                transformer_model = TFAutoModel.from_pretrained(self.local_dir_path, from_pt=True)
                self.transformer_config = transformer_model.config
                return transformer_model
            else:
                config_path = os.path.join(".", self.local_dir_path, TRANSFORMER_CONFIG_FILE_NAME)
                self.transformer_config = AutoConfig.from_pretrained(config_path)
                return TFAutoModel.from_config(self.transformer_config)



import numpy as np
import regex as re
from unidecode import unidecode

from delft.utilities.Tokenizer import tokenizeAndFilterSimple

special_character_removal = re.compile(r"[^A-Za-z\.\-\?\!\,\#\@\% ]", re.IGNORECASE)


def to_vector_single(text, embeddings, maxlen=300):
    """
    Given a string, tokenize it, then convert it to a sequence of word embedding
    vectors with the provided embeddings, introducing <PAD> and <UNK> padding token
    vector when appropriate
    """
    tokens = tokenizeAndFilterSimple(clean_text(text))
    window = tokens[-maxlen:]

    # TBD: use better initializers (uniform, etc.)
    x = np.zeros(
        (maxlen, embeddings.embed_size),
    )

    # TBD: padding should be left and which vector do we use for padding?
    # and what about masking padding later for RNN?
    for i, word in enumerate(window):
        x[i, :] = embeddings.get_word_vector(word).astype("float32")

    return x


def clean_text(text):
    x_ascii = unidecode(text)
    x_clean = special_character_removal.sub("", x_ascii)
    return x_clean


def lower(word):
    return word.lower()


def normalize_num(word):
    return re.sub(r"[0-9０１２３４５６７８９]", r"0", word)


def create_single_input_bert(text, maxlen=512, transformer_tokenizer=None):
    """
    Note: use batch method preferably for better performance
    """

    # TBD: exception if tokenizer is not valid/None
    encoded_tokens = transformer_tokenizer.encode_plus(
        text, truncation=True, add_special_tokens=True, max_length=maxlen, padding="max_length"
    )
    # note: [CLS] and [SEP] are added by the tokenizer

    ids = encoded_tokens["input_ids"]
    masks = encoded_tokens["token_type_ids"]
    segments = encoded_tokens["attention_mask"]

    return ids, masks, segments


def create_batch_input_bert(texts, maxlen=512, transformer_tokenizer=None):
    # TBD: exception if tokenizer is not valid/None

    if isinstance(texts, np.ndarray):
        texts = texts.tolist()

    encoded_tokens = transformer_tokenizer.batch_encode_plus(
        texts, add_special_tokens=True, truncation=True, max_length=maxlen, padding="max_length"
    )

    # note: special tokens like [CLS] and [SEP] are added by the tokenizer

    ids = encoded_tokens["input_ids"]
    masks = encoded_tokens["token_type_ids"]
    segments = encoded_tokens["attention_mask"]

    return ids, masks, segments



import logging
import os

import numpy as np

# derived from https://github.com/elifesciences/sciencebeam-trainer-delft/tree/develop/tests
from delft.sequenceLabelling.preprocess import FeaturesPreprocessor, Preprocessor

LOGGER = logging.getLogger(__name__)

FEATURE_VALUE_1 = "feature1"
FEATURE_VALUE_2 = "feature2"
FEATURE_VALUE_3 = "feature3"
FEATURE_VALUE_4 = "feature4"


class TestWordPreprocessor:
    def test_should_be_able_to_instantiate_with_default_values(self):
        Preprocessor()

    def test_should_fit_empty_dataset(self):
        preprocessor = Preprocessor()
        preprocessor.fit([], [])

    def test_should_fit_single_word_dataset(self):
        preprocessor = Preprocessor()
        X = [["Word1"]]
        y = [["label1"]]
        X_transformed, y_transformed = preprocessor.fit_transform(X, y)
        LOGGER.debug("vocab_char: %s", preprocessor.vocab_char)
        LOGGER.debug("vocab_case: %s", preprocessor.vocab_case)
        LOGGER.debug("vocab_tag: %s", preprocessor.vocab_tag)
        LOGGER.debug("X_transformed: %s", X_transformed)
        LOGGER.debug("y_transformed: %s", y_transformed)
        for c in "Word1":
            assert c in preprocessor.vocab_char
        for case in {"numeric", "allLower", "allUpper", "initialUpper"}:
            assert case in preprocessor.vocab_case
        assert "label1" in preprocessor.vocab_tag
        assert len(X_transformed) == 1
        assert len(y_transformed) == 1

    def test_should_be_able_to_inverse_transform_label(self):
        preprocessor = Preprocessor()
        X = [["Word1"]]
        y = [["label1"]]
        _, y_transformed = preprocessor.fit_transform(X, y)
        y_inverse = preprocessor.inverse_transform(y_transformed[0])
        assert y_inverse == y[0]

    def test_should_transform_unseen_label(self):
        preprocessor = Preprocessor(return_lengths=False, padding=False)
        X_train = [["Word1"]]
        y_train = [["label1"]]
        X_test = [["Word1", "Word1"]]
        y_test = [["label1", "label2"]]
        p = preprocessor.fit(X_train, y_train)
        _, y_transformed = p.transform(X_test, y_test)
        assert y_transformed == [[1, 0]]

    def test_load_example(self, preprocessor1):
        p = Preprocessor.load(preprocessor1)

        assert len(p.vocab_char) == 70

    def test_load_withUmmappedVariable_shouldIgnore(self, preprocessor2: str):
        p = Preprocessor.load(preprocessor2)

        assert len(p.vocab_char) == 70


def _to_dense(a: np.array):
    try:
        return a.todense()
    except AttributeError:
        return a


def all_close(a: np.array, b: np.array):
    return np.allclose(_to_dense(a), _to_dense(b))


class TestFeaturesPreprocessor:
    def test_should_be_able_to_instantiate_with_default_values(self):
        FeaturesPreprocessor()

    def test_should_fit_empty_dataset(self):
        preprocessor = FeaturesPreprocessor()
        preprocessor.fit([])

    def test_should_fit_single_value_feature(self):
        preprocessor = FeaturesPreprocessor()
        features_batch = [[[FEATURE_VALUE_1]]]
        features_transformed = preprocessor.fit_transform(features_batch)
        features_length = len(preprocessor.features_indices)
        assert features_length == 1
        assert all_close(features_transformed, [[[1]]])

    def test_should_fit_single_multiple_value_features(self):
        preprocessor = FeaturesPreprocessor()
        features_batch = [[[FEATURE_VALUE_1], [FEATURE_VALUE_2]]]
        features_transformed = preprocessor.fit_transform(features_batch)
        features_length = len(preprocessor.features_indices)
        assert features_length == 1
        assert len(features_transformed[0]) == 2
        assert np.array_equal(features_transformed, np.asarray([[[1], [2]]]))

    def test_should_fit_multiple_single_value_features(self):
        preprocessor = FeaturesPreprocessor()
        features_batch = [[[FEATURE_VALUE_1, FEATURE_VALUE_2]]]
        features_transformed = preprocessor.fit_transform(features_batch)
        features_length = len(preprocessor.features_indices)
        assert features_length == 2
        assert all_close(features_transformed, [[[1, 13]]])

    def test_should_transform_unseen_to_zero(self):
        preprocessor = FeaturesPreprocessor()
        features_batch = [[[FEATURE_VALUE_1]]]
        preprocessor.fit(features_batch)
        features_transformed = preprocessor.transform([[[FEATURE_VALUE_2]]])
        assert all_close(features_transformed, [[[0]]])

    def test_should_select_features(self):
        preprocessor = FeaturesPreprocessor(features_indices=[1])
        features_batch = [
            [[FEATURE_VALUE_1, FEATURE_VALUE_2], [FEATURE_VALUE_1, FEATURE_VALUE_3], [FEATURE_VALUE_1, FEATURE_VALUE_4]]
        ]
        features_transformed = preprocessor.fit_transform(features_batch)
        features_length = len(preprocessor.features_indices)
        assert features_length == 1
        assert all_close(features_transformed, [[[1], [2], [3]]])

    def test_serialize_to_json(self, tmp_path):
        preprocessor = FeaturesPreprocessor(features_indices=[1])
        features_batch = [
            [[FEATURE_VALUE_1, FEATURE_VALUE_2], [FEATURE_VALUE_1, FEATURE_VALUE_3], [FEATURE_VALUE_1, FEATURE_VALUE_4]]
        ]
        X_train = [["Word1"]]
        y_train = [["label1"]]
        preprocessor.fit(features_batch)
        word_preprocessor = Preprocessor(feature_preprocessor=preprocessor)
        word_preprocessor.fit(X_train, y_train)

        serialised_file_path = os.path.join(str(tmp_path), "serialised.json")
        word_preprocessor.save(file_path=serialised_file_path)

        back = Preprocessor.load(serialised_file_path)

        assert back is not None
        assert back.feature_preprocessor is not None
        original_as_dict = word_preprocessor.__dict__
        back_as_dict = back.__dict__
        for key in back_as_dict.keys():
            if key == "feature_preprocessor":
                for sub_key in back_as_dict[key].__dict__.keys():
                    assert back_as_dict[key].__dict__[sub_key] == original_as_dict[key].__dict__[sub_key]
            else:
                assert back_as_dict[key] == original_as_dict[key]
